from __future__ import annotations

from copy import deepcopy


def merge_dict(d1, d2):
    """
    Merge two nested dictionaries together.

    Nested dictionaries are not overwritten but combined.
    """
    dest = deepcopy(d1)
    for key, value in d2.items():
        if isinstance(value, dict):
            dest[key] = merge_dict(dest.get(key, {}), value)
            continue

        dest[key] = value

    return dest



from __future__ import annotations

import pytest

from .marks import TopologyMark
from .multihost import MultihostConfig


class MultihostItemData(object):
    """
    Multihost internal pytest data, stored in :attr:`pytest.Item.multihost`
    """

    def __init__(self, multihost: MultihostConfig | None, topology_mark: TopologyMark | None) -> None:
        self.multihost: MultihostConfig | None = multihost
        """
        Multihost object.
        """

        self.topology_mark: TopologyMark | None = topology_mark
        """
        Topology mark for the test run.
        """

        self.outcome: str | None = None
        """
        Test run outcome, available in fixture finalizers.
        """

    @staticmethod
    def SetData(item: pytest.Item, data: MultihostItemData | None) -> None:
        item.stash[DataStashKey] = data

    @staticmethod
    def GetData(item: pytest.Item) -> MultihostItemData | None:
        return item.stash[DataStashKey]


DataStashKey = pytest.StashKey[MultihostItemData | None]()



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from base64 import b64decode
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Type, TypeVar

from ..cli import CLIBuilder
from ..ssh import SSHBashProcess, SSHClient, SSHLog, SSHPowerShellProcess, SSHProcess
from .logging import MultihostHostLogger, MultihostLogger
from .marks import TopologyMark
from .utils import validate_configuration

if TYPE_CHECKING:
    from .fixtures import MultihostFixture


class MultihostConfig(ABC):
    """
    Multihost configuration.
    """

    def __init__(self, confdict: dict[str, Any], *, logger: MultihostLogger, lazy_ssh: bool = False) -> None:
        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in configuration'
        )

        self.logger: MultihostLogger = logger
        """Multihost logger"""

        self.lazy_ssh: bool = lazy_ssh
        """If True, hosts postpone connecting to ssh when the connection is first required"""

        self.domains: list[MultihostDomain] = []
        """Available domains"""

        for domain in confdict["domains"]:
            self.domains.append(self.create_domain(domain))

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the host configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["domains"]

    @property
    def TopologyMarkClass(self) -> Type[TopologyMark]:
        """
        Class name of the type or subtype of :class:`TopologyMark`.
        """
        return TopologyMark

    def create_domain(self, domain: dict[str, Any]) -> MultihostDomain:
        """
        Create new multihost domain from dictionary.

        It maps the role name to a Python class using
        :attr:`id_to_domain_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it raises
        ``ValueError``.

        :param domain: Domain in dictionary form.
        :type domain: dict[str, Any]
        :raises ValueError: If domain does not have id or mapping to Python class is not found.
        :return: New multihost domain.
        :rtype: MultihostDomain
        """
        id = domain.get("id", None)
        if id is None:
            raise ValueError("Invalid configuration, domain is missing 'id'")

        cls = self.id_to_domain_class.get(id, self.id_to_domain_class.get("*", None))
        if cls is None:
            raise ValueError(f"Unexpected domain id: {id}")

        return cls(self, domain)

    @property
    @abstractmethod
    def id_to_domain_class(self) -> dict[str, Type[MultihostDomain]]:
        """
        Map domain id to domain class. Asterisk ``*`` can be used as fallback
        value.

        :rtype: Class name.
        """
        pass


ConfigType = TypeVar("ConfigType", bound=MultihostConfig)


class MultihostDomain(ABC, Generic[ConfigType]):
    """
    Multihost domain class.
    """

    def __init__(self, config: ConfigType, confdict: dict[str, Any]) -> None:
        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in domain configuration'
        )

        self.config: ConfigType = config
        """Multihost configuration"""

        self.logger: MultihostLogger = config.logger
        """Multihost logger"""

        self.id: str = confdict["id"]
        """Domain id"""

        self.hosts: list[MultihostHost] = []
        """Available hosts in this domain"""

        for host in confdict["hosts"]:
            self.hosts.append(self.create_host(host))

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the domain configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["id", "hosts"]

    @property
    def roles(self) -> list[str]:
        """
        All roles available in this domain.

        :return: Role names.
        :rtype: list[str]
        """
        return sorted(set(x.role for x in self.hosts))

    def create_host(self, confdict: dict[str, Any]) -> MultihostHost:
        """
        Create host object from role.

        It maps the role name to a Python class using
        :attr:`role_to_host_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it fallbacks to
        :class:`MultiHost`.

        :param confdict: Host configuration as a dictionary.
        :type confdict: dict[str, Any]
        :raises ValueError: If role property is missing in the host
            configuration.
        :return: Host instance.
        :rtype: MultihostHost
        """
        if not confdict.get("role", None):
            raise ValueError('"role" property is missing in host configuration')

        role = confdict["role"]
        cls = self.role_to_host_class.get(role, self.role_to_host_class.get("*", MultihostHost))

        return cls(self, confdict)

    def create_role(self, mh: MultihostFixture, host: MultihostHost) -> MultihostRole:
        """
        Create role object from given host.

        It maps the role name to a Python class using
        :attr:`role_to_role_class`. If the role is not found in the property, it
        fallbacks to ``*``. If even asterisk is not found, it raises
        ``ValueError``.

        :param mh: Multihost instance.
        :type mh: Multihost
        :param host: Multihost host instance.
        :type host: MultihostHost
        :raises ValueError: If unexpected role name is given.
        :return: Role instance.
        :rtype: MultihostRole
        """
        cls = self.role_to_role_class.get(host.role, self.role_to_role_class.get("*", None))
        if cls is None:
            raise ValueError(f"Unexpected role: {host.role}")

        return cls(mh, host.role, host)

    @property
    @abstractmethod
    def role_to_host_class(self) -> dict[str, Type[MultihostHost]]:
        """
        Map role to host class. Asterisk ``*`` can be used as fallback value.

        :rtype: Class name.
        """
        pass

    @property
    @abstractmethod
    def role_to_role_class(self) -> dict[str, Type[MultihostRole]]:
        """
        Map role to role class. Asterisk ``*`` can be used as fallback value.

        :rtype: Class name.
        """
        pass

    def hosts_by_role(self, role: str) -> list[MultihostHost]:
        """
        Return all hosts of the given role.

        :param role: Role name.
        :type role: str
        :return: List of hosts of given role.
        :rtype: list[MultihostHost]
        """
        return [x for x in self.hosts if x.role == role]


DomainType = TypeVar("DomainType", bound=MultihostDomain)


class MultihostHostOSFamily(Enum):
    """
    Host operating system family.
    """

    Linux = "linux"
    Windows = "windows"


class MultihostHost(Generic[DomainType]):
    """
    Base multihost host class.

    .. code-block:: yaml
        :caption: Example configuration in YAML format

        - hostname: dc.ad.test
          role: ad
          os:
            family: linux
          ssh:
            host: 1.2.3.4
            username: root
            password: Secret123
          config:
            binddn: Administrator@ad.test
            bindpw: vagrant
            client:
              ad_domain: ad.test
              krb5_keytab: /enrollment/ad.keytab
              ldap_krb5_keytab: /enrollment/ad.keytab

    * Required fields: ``hostname``, ``role``
    * Optional fields: ``artifacts``, ``config``, ``os``, ``ssh``
    """

    def __init__(self, domain: DomainType, confdict: dict[str, Any]):
        """
        :param domain: Multihost domain object.
        :type domain: DomainType
        :param confdict: Host configuration as a dictionary.
        :type confdict: dict[str, Any]
        :param shell: Shell used in SSH connection, defaults to '/usr/bin/bash -c'.
        :type shell: str
        """
        validate_configuration(
            self.required_fields, confdict, error_fmt='"{key}" property is missing in host configuration'
        )

        # Required
        self.domain: DomainType = domain
        """Multihost domain."""

        self.role: str = confdict["role"]
        """Host role."""

        self.hostname: str = confdict["hostname"]
        """Host hostname."""

        self.logger: MultihostLogger = self.domain.logger.subclass(
            cls=MultihostHostLogger, suffix=f"host.{self.hostname}", hostname=self.hostname
        )
        """Multihost logger."""

        # Optional
        self.config: dict[str, Any] = confdict.get("config", {})
        """Custom configuration."""

        self.artifacts: list[str] = confdict.get("artifacts", [])
        """Host artifacts produced during tests."""

        # SSH
        ssh = confdict.get("ssh", {})

        self.ssh_host: str = ssh.get("host", self.hostname)
        """SSH host (resolvable hostname or IP address), defaults to :attr:`hostname`."""

        self.ssh_port: int = int(ssh.get("port", 22))
        """SSH port, defaults to ``22``."""

        self.ssh_username: str = ssh.get("username", "root")
        """SSH username, defaults to ``root``."""

        self.ssh_password: str = ssh.get("password", "Secret123")
        """SSH password, defaults to ``Secret123``."""

        # Not configurable
        self.shell: Type[SSHProcess] = SSHBashProcess
        """Shell used in SSH session."""

        # Get host operating system information
        os = confdict.get("os", {})

        os_family = str(os.get("family", MultihostHostOSFamily.Linux.value)).lower()
        try:
            self.os_family: MultihostHostOSFamily = MultihostHostOSFamily(os_family)
            """Host operating system os_family."""
        except ValueError:
            raise ValueError(f'Value "{os_family}" is not supported in os_family field of host configuration')

        # Set host shell based on the operating system
        match self.os_family:
            case MultihostHostOSFamily.Linux:
                self.shell = SSHBashProcess
            case MultihostHostOSFamily.Windows:
                self.shell = SSHPowerShellProcess
            case _:
                raise ValueError(f"Unknown operating system os_family: {self.os_family}")

        # SSH connection
        self.ssh: SSHClient = SSHClient(
            host=self.ssh_host,
            user=self.ssh_username,
            password=self.ssh_password,
            port=self.ssh_port,
            logger=self.logger,
            shell=self.shell,
        )
        """SSH client."""

        # CLI Builder instance
        self.cli: CLIBuilder = CLIBuilder(self.ssh)
        """Command line builder."""

        # Connect to SSH unless lazy ssh is set
        if not self.domain.config.lazy_ssh:
            self.ssh.connect()

    @property
    def required_fields(self) -> list[str]:
        """
        Fields that must be set in the host configuration. An error is raised
        if any field is missing.

        The field name may contain a ``.`` to check nested fields.
        """
        return ["role", "hostname"]

    def collect_artifacts(self, dest: str, additional_artifacts: list[str]) -> None:
        """
        Collect test artifacts that were requested by the multihost
        configuration.

        :param dest: Destination directory, where the artifacts will be stored.
        :type dest: str
        :param additional_artifacts: Additional artifacts that will be fetched
            together with artifacts from configuration file.
        :type additional_artifacts: list[str]
        """
        artifacts = sorted(list(set(self.artifacts + additional_artifacts)))
        if not artifacts:
            return

        self.logger.info(
            "Collecting artifacts",
            extra={
                "data": {
                    "Local destination": dest,
                    "Artifacts": artifacts,
                }
            },
        )

        # Create output directory
        Path(dest).mkdir(parents=True, exist_ok=True)

        # Fetch artifacts
        match self.os_family:
            case MultihostHostOSFamily.Linux:
                command = f"""
                    tmp=`mktemp /tmp/mh.host.artifacts.XXXXXXXXX`
                    tar -czvf "$tmp" {' '.join([f'$(compgen -G "{x}")' for x in artifacts])} &> /dev/null
                    base64 "$tmp"
                    rm -f "$tmp" &> /dev/null
                """
                ext = "tgz"
            case MultihostHostOSFamily.Windows:
                raise NotImplementedError("Artifacts are not supported on Windows machine")
            case _:
                raise ValueError(f"Unknown operating system: {self.os_family}")

        result = self.ssh.run(command, log_level=SSHLog.Error)

        # Store artifacts in single archive
        with open(f"{dest}/{self.role}_{self.hostname}.{ext}", "wb") as f:
            f.write(b64decode(result.stdout))

    def pytest_setup(self) -> None:
        """
        Called once before execution of any tests.
        """
        pass

    def pytest_teardown(self) -> None:
        """
        Called once after all tests are finished.
        """
        pass

    def setup(self) -> None:
        """
        Called before execution of each test.
        """
        pass

    def teardown(self) -> None:
        """
        Called after execution of each test.
        """
        pass


HostType = TypeVar("HostType", bound=MultihostHost)


class MultihostRole(Generic[HostType]):
    """
    Base role class. Roles are the main interface to the remote hosts that can
    be directly accessed in test cases as fixtures.

    All changes to the remote host that were done through the role object API
    are automatically reverted when a test is finished.
    """

    def __init__(
        self,
        mh: MultihostFixture,
        role: str,
        host: HostType,
    ) -> None:
        self.mh: MultihostFixture = mh
        self.role: str = role
        self.host: HostType = host

        self.logger: MultihostLogger = self.host.logger
        """Multihost logger."""

        self.artifacts: list[str] = []
        """
        List of artifacts that will be automatically collected when a test is
        finished. This list can be dynamically extended. Values may contain
        wildcard character.
        """

    def setup(self) -> None:
        """
        Setup all :class:`MultihostUtility` objects
        that are attributes of this class.
        """
        MultihostUtility.SetupUtilityAttributes(self)

    def teardown(self) -> None:
        """
        Teardown all :class:`MultihostUtility` objects
        that are attributes of this class.
        """
        MultihostUtility.TeardownUtilityAttributes(self)

    def ssh(self, user: str, password: str, *, shell=SSHBashProcess) -> SSHClient:
        """
        Open SSH connection to the host as given user.

        :param user: Username.
        :type user: str
        :param password: User password.
        :type password: str
        :param shell: Shell that will run the commands, defaults to SSHBashProcess
        :type shell: str, optional
        :return: SSH client connection.
        :rtype: SSHClient
        """
        return SSHClient(
            self.host.ssh_host,
            user=user,
            password=password,
            port=self.host.ssh_port,
            shell=shell,
            logger=self.logger,
        )


class MultihostUtility(Generic[HostType]):
    """
    Base class for utility functions that operate on remote hosts, such as
    writing a file or managing SSSD.

    Instances of :class:`MultihostUtility` can be used in any role class which
    is a subclass of :class:`MultihostRole`. In this case, :func:`setup` and
    :func:`teardown` methods are called automatically when the object is created
    and destroyed to ensure proper setup and clean up on the remote host.
    """

    def __init__(self, host: HostType) -> None:
        """
        :param host: Remote host instance.
        :type host: HostType
        """
        self.host: HostType = host
        """Multihost host."""

        self.logger: MultihostLogger = self.host.logger
        """Multihost logger."""

        self.used: bool = False
        """Indicate if this utility instance was already used or not within current test."""

        # Enable first use setup
        disallowed = [
            "setup",
            "teardown",
            "setup_when_used",
            "teardown_when_used",
            "GetUtilityAttributes",
            "SetupUtilityAttributes",
            "TeardownUtilityAttributes",
        ]
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if name in disallowed or name.startswith("_") or hasattr(method, "__mh_ignore_call"):
                continue

            setattr(self, name, self.__setup_when_used(method))

    def setup(self) -> None:
        """
        Setup object.
        """
        pass

    def teardown(self) -> None:
        """
        Teardown object.
        """
        pass

    def setup_when_used(self) -> None:
        """
        Setup the object when it is used for the first time.
        """
        pass

    def teardown_when_used(self) -> None:
        """
        Teardown the object only if it was used.
        """
        pass

    def __setup_when_used(self, method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            if not self.used:
                self.used = True
                self.setup_when_used()

            self.used = True
            return method(*args, **kwargs)

        return wrapper

    @staticmethod
    def GetUtilityAttributes(o: object) -> dict[str, MultihostUtility]:
        """
        Get all attributes of the ``o`` that are instance of
        :class:`MultihostUtility`.

        :param o: Any object.
        :type o: object
        :return: Dictionary {attribute name: value}
        :rtype: dict[str, MultihostUtility]
        """
        return dict(inspect.getmembers(o, lambda attr: isinstance(attr, MultihostUtility)))

    @classmethod
    def SetupUtilityAttributes(cls, o: object) -> None:
        """
        Setup all :class:`MultihostUtility` objects attributes of the given
        object.

        :param o: Any object.
        :type o: object
        """
        for util in cls.GetUtilityAttributes(o).values():
            util.setup()

    @classmethod
    def TeardownUtilityAttributes(cls, o: object) -> None:
        """
        Teardown all :class:`MultihostUtility` objects attributes of the given
        object.

        :param o: Any object.
        :type o: object
        """
        errors = []
        for util in cls.GetUtilityAttributes(o).values():
            if util.used:
                try:
                    util.teardown_when_used()
                except Exception as e:
                    errors.append(e)

            try:
                util.teardown()
            except Exception as e:
                errors.append(e)

        if errors:
            raise Exception(errors)

    @classmethod
    def IgnoreCall(cls, method):
        """
        Calling a method decorated with IgnoreCall does not execute neither
        :meth:`setup_when_used` nor :meth:`teardown_when_used`. It does not
        count as "using" the class.
        """
        method.__mh_ignore_call = True
        return method



from __future__ import annotations

from typing import Any

from pytest_mh.cli import CLIBuilder, CLIBuilderArgs

from .. import MultihostHost, MultihostUtility
from ..ssh import SSHLog, SSHProcessResult

__all__ = ["JournaldUtils"]


class JournaldUtils(MultihostUtility):
    """
    Perform journald related tasks.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)
        self._test_start: str = ""

    def setup(self) -> None:
        """
        Called before execution of each test.
        """
        self.clear()

    def clear(self) -> None:
        """
        Reset timestamp
        """
        self._test_start = self.host.ssh.exec(["date", "--rfc-3339=ns"], log_level=SSHLog.Error).stdout[0:19]

    def journalctl(
        self,
        current: bool = True,
        *,
        unit: str | None = None,
        lines: int | None = None,
        since: str | None = None,
        reverse: bool = False,
        no_pager: bool = False,
        grep: str | None = None,
        output: str | None = None,
        identifier: str | None = None,
        system: bool = False,
        user: bool = False,
        args: list[Any] | None = None,
    ) -> SSHProcessResult:
        """
        Execute journalctl with given arguments. Show messages only for current test run, by default.
        Note that raise_on_error is False and the command may return non-zero return code.

        :param current: Show messages only for current test run, defaults to True
        :type current: bool, optional
        :param unit: Show messages for the specified systemd unit, defaults to None
        :type unit: str | None, optional
        :param lines: Show the most recent journal events and limit the number of events shown, defaults to None
        :type lines: int | None, optional
        :param since: Start showing entries on or newer than the specified date, defaults to None
        :type since: str | None, optional
        :param reverse: Reverse output so that the newest entries are displayed first, defaults to False
        :type reverse: bool, optional
        :param no_pager: Do not pipe output into a pager, defaults to False
        :type no_pager: bool, optional
        :param grep: Filter output to entries where the MESSAGE= field matches specified regex, defaults to None
        :type grep: str | None, optional
        :param output: Controls the formatting of the journal entries, defaults to None
        :type output: str | None, optional
        :param identifier: Show messages for the specified syslog identifier SYSLOG_IDENTIFIER, defaults to None
        :type identifier: str | None, optional
        :param system: Show messages from system services and the kernel, defaults to False
        :type system: bool, optional
        :param user: Show messages from service of current user, defaults to False
        :type user: bool, optional
        :param args: Additional options, defaults to None
        :type args: list[Any] | None, optional
        :return: SSH process result
        :rtype: SSHProcessResult
        """
        cli: CLIBuilder = CLIBuilder(self.host.ssh)
        if current:
            since = since if since else self._test_start

        args = args if args else []
        builder: CLIBuilderArgs = {
            "unit": (cli.option.VALUE, unit),
            "lines": (cli.option.VALUE, lines),
            "since": (cli.option.VALUE, since),
            "reverse": (cli.option.SWITCH, reverse),
            "no-pager": (cli.option.SWITCH, no_pager),
            "grep": (cli.option.VALUE, grep),
            "output": (cli.option.VALUE, output),
            "identifier": (cli.option.VALUE, identifier),
            "system": (cli.option.SWITCH, system),
            "user": (cli.option.SWITCH, user),
        }

        return self.host.ssh.exec(["journalctl"] + cli.args(builder) + args, raise_on_error=False)

    def is_match(self, pattern: str) -> bool:
        """
        Search the logs for a pattern.

        :param pattern: Pattern to be searched for
        :type pattern: str
        :return: True, if pattern found
        :rtype: bool
        """
        return self.journalctl(grep=pattern).rc == 0

    def count(self, pattern: str) -> int:
        """
        Search the logs for a pattern and return number of occurrences.

        :param pattern: Pattern to be searched for
        :type pattern: str
        :return: Number of occurrences of the pattern
        :rtype: int
        """
        process = self.journalctl(grep=pattern)
        return len(process.stdout_lines) if process.rc == 0 else 0



"""
.. avoid already-imported warning: PYTEST_DONT_REWRITE (hidden from sphinx)
"""
from __future__ import annotations

from ._private.data import MultihostItemData
from ._private.fixtures import MultihostFixture, mh
from ._private.marks import KnownTopologyBase, KnownTopologyGroupBase, TopologyMark
from ._private.multihost import (
    MultihostConfig,
    MultihostDomain,
    MultihostHost,
    MultihostHostOSFamily,
    MultihostRole,
    MultihostUtility,
)
from ._private.plugin import MultihostPlugin, pytest_addoption, pytest_configure
from ._private.topology import Topology, TopologyDomain

__all__ = [
    "mh",
    "MultihostConfig",
    "MultihostDomain",
    "MultihostFixture",
    "MultihostHost",
    "MultihostHostOSFamily",
    "MultihostItemData",
    "MultihostPlugin",
    "MultihostRole",
    "MultihostUtility",
    "pytest_addoption",
    "pytest_configure",
    "Topology",
    "TopologyDomain",
    "TopologyMark",
    "KnownTopologyBase",
    "KnownTopologyGroupBase",
]



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Type, TypeAlias

from .ssh import SSHClient, SSHPowerShellProcess, SSHProcess


class CLIBuilder(object):
    class option(Enum):
        """
        Command line parameter types.
        """

        PLAIN = auto()
        """
        Use plain parameter value without any modification.
        """

        VALUE = auto()
        """
        Use parameter value but enclose it in quotes in script mode.
        """

        SWITCH = auto()
        """
        Parameter is a switch which is enabled if value is True.
        """

        POSITIONAL = auto()
        """
        Parameter is a positional argument.
        """

    def __init__(self, ssh: SSHClient) -> None:
        self.__shell: Type[SSHProcess] = ssh.shell
        self.__prefix: str = "-" if self.__match_shell(SSHPowerShellProcess) else "--"

    def command(self, command: str, args: CLIBuilderArgs) -> str:
        return " ".join(self.__build(command, args, quote_value=True))

    def argv(self, command: str, args: CLIBuilderArgs) -> list[str]:
        return self.__build(command, args, quote_value=False)

    def args(self, args: CLIBuilderArgs, quote_value=False) -> list[str]:
        return self.__build(None, args, quote_value)

    def __match_shell(self, shell: Type[SSHProcess]):
        return issubclass(self.__shell, shell)

    def __build(self, command: str | None, args: CLIBuilderArgs, quote_value: bool) -> list[str]:
        def _get_option(name: str) -> str:
            return self.__prefix + name

        def _get_value(value: Any) -> str:
            return str(value) if not quote_value else f"'{value}'"

        argv = [command] if command is not None else []
        for key, item in args.items():
            if item is None:
                continue

            (type, value) = item
            if value is None:
                continue

            match type:
                case self.option.POSITIONAL:
                    argv.append(_get_value(value))
                case self.option.SWITCH:
                    if self.__match_shell(SSHPowerShellProcess):
                        argv.append(f'{_get_option(key)}:{"$True" if value else "$False"}')
                    else:
                        if value:
                            argv.append(_get_option(key))
                case self.option.VALUE:
                    argv.append(_get_option(key))
                    argv.append(_get_value(value))
                case self.option.PLAIN:
                    argv.append(_get_option(key))
                    argv.append(str(value))
                case _:
                    raise ValueError(f"Unknown option type: {type}")

        return argv


CLIBuilderArgs: TypeAlias = dict[str, tuple[CLIBuilder.option, Any] | None]
"""CLIBuilder args format."""



from __future__ import annotations

import inspect
import logging
import sys
import textwrap
from typing import Generator, Type

import pytest
import yaml

from .data import MultihostItemData
from .fixtures import MultihostFixture
from .logging import MultihostLogger
from .marks import TopologyMark
from .multihost import MultihostConfig, MultihostHost
from .topology import Topology

MarkStashKey = pytest.StashKey[TopologyMark | None]()


class MultihostPlugin(object):
    """
    Pytest multihost plugin.
    """

    def __init__(self, pytest_config: pytest.Config) -> None:
        self.config_class: Type[MultihostConfig] | None = None
        self.logger: logging.Logger = self._create_logger(pytest_config.option.verbose > 2)
        self.multihost: MultihostConfig | None = None
        self.topology: Topology | None = None
        self.confdict: dict | None = None

        # CLI options
        self.mh_config: str = pytest_config.getoption("mh_config")
        self.mh_log_path: str = pytest_config.getoption("mh_log_path")
        self.mh_lazy_ssh: bool = pytest_config.getoption("mh_lazy_ssh")
        self.mh_topology: list[str] = pytest_config.getoption("mh_topology")
        self.mh_not_topology: list[str] = pytest_config.getoption("mh_not_topology")
        self.mh_exact_topology: bool = pytest_config.getoption("mh_exact_topology")
        self.mh_collect_artifacts: bool = pytest_config.getoption("mh_collect_artifacts")
        self.mh_artifacts_dir: bool = pytest_config.getoption("mh_artifacts_dir")

    @classmethod
    def GetLogger(cls) -> logging.Logger:
        """
        Get plugin's logger.
        """

        return logging.getLogger("pytest_mh.plugin")

    def __load_conf(self, path: str) -> dict:
        """
        Load multihost configuration from a yaml file.

        :param path: Path to the yaml file.
        :type path: str
        :raises ValueError: If not file was provided.
        :raises IOError: If unable to read the file.
        :return: Parsed configuration.
        :rtype: dict

        :meta private:
        """
        if not path:
            raise ValueError("You need to provide valid multihost configuration file, use --mh-config=$path")

        try:
            with open(path, "r") as f:
                return yaml.safe_load(f)
        except Exception as e:
            raise IOError(f'Unable to open multihost configuration "{path}": {str(e)}')

    def setup(self) -> None:
        """
        Read and apply multihost configuration.

        :meta private:
        """
        if self.config_class is None:
            raise ValueError("Set MultihostPlugin.config_class to subclass of MultihostConfig")

        self.confdict = self.__load_conf(self.mh_config)

        logger = MultihostLogger.GetLogger()
        logger.setup(log_path=self.mh_log_path, confdict=self.confdict)

        self.multihost = self.config_class(self.confdict, logger=logger, lazy_ssh=self.mh_lazy_ssh)
        self.topology = Topology.FromMultihostConfig(self.confdict)

    @pytest.hookimpl(trylast=True)
    def pytest_sessionstart(self, session: pytest.Session) -> None:
        """
        Setup the module and log information about given multihost configuration
        and provided options.

        :meta private:
        """
        # Calling the setup here instead of in constructor to allow running
        # pytest --help and other action-less parameters.
        self.setup()

        # Silent mypy false positive
        if self.topology is None:
            raise ValueError("Topology must be already set!")

        if self.multihost is None:
            self.logger.info(self._fmt_bold("Multihost configuration:"))
            self.logger.info("  No multihost configuration provided.")
            self.logger.info("  Make sure to run tests with --mh-log-path parameter.")
            self.logger.info("")
            return

        self.logger.info(self._fmt_bold("Multihost configuration:"))
        self.logger.info(textwrap.indent(yaml.dump(self.confdict, sort_keys=False), "  "))
        self.logger.info(self._fmt_bold("Detected topology:"))
        self.logger.info(textwrap.indent(yaml.dump(self.topology.export(), sort_keys=False), "  "))
        self.logger.info(self._fmt_bold("Additional settings:"))
        self.logger.info(f"  config file: {self.mh_config}")
        self.logger.info(f"  log path: {self.mh_log_path}")
        self.logger.info(f"  lazy ssh: {self.mh_lazy_ssh}")
        self.logger.info(f"  topology filter: {', '.join(self.mh_topology + [f'!{x}' for x in self.mh_not_topology])}")
        self.logger.info(f"  require exact topology: {self.mh_exact_topology}")
        self.logger.info(f"  collect artifacts: {self.mh_collect_artifacts}")
        self.logger.info(f"  artifacts directory: {self.mh_artifacts_dir}")
        self.logger.info("")

        try:
            setup_ok: list[MultihostHost] = []
            for domain in self.multihost.domains:
                for host in domain.hosts:
                    try:
                        host.pytest_setup()
                    except Exception:
                        # Teardown hosts that were successfully setup before this error
                        for h in reversed(setup_ok):
                            h.pytest_teardown()
                        raise

                    setup_ok.append(host)
        finally:
            self.multihost.logger.write_to_file(f"{self.mh_artifacts_dir}/setup.log")

    @pytest.hookimpl(trylast=True)
    def pytest_sessionfinish(self, session: pytest.Session, exitstatus: int | pytest.ExitCode) -> None:
        """
        Teardown the module.

        :meta private:
        """
        if self.multihost is None:
            return

        errors = []
        for domain in self.multihost.domains:
            for host in domain.hosts:
                try:
                    host.pytest_teardown()
                except Exception as e:
                    errors.append(e)

        self.multihost.logger.write_to_file(f"{self.mh_artifacts_dir}/teardown.log")

        if errors:
            raise Exception(errors)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_make_collect_report(self, collector: pytest.Collector) -> Generator[None, pytest.CollectReport, None]:
        """
        If multiple topology marks are present on the collected test, we need to
        parametrize it. In order to do so, the test is replaced with multiple
        clones, one for each topology.

        The topology associated with the test clone is stored in
        ``topology_mark`` property of the clone.

        :meta private:
        """

        outcome = yield
        report = outcome.get_result()

        if not report.result:
            return

        if self.multihost is None:
            return

        new_result = []
        for result in report.result:
            if not isinstance(result, pytest.Function):
                new_result.append(result)
                continue

            has_marks = False
            for mark in self.multihost.TopologyMarkClass.ExpandMarkers(result):
                has_marks = True
                topology_mark = self.multihost.TopologyMarkClass.Create(result, mark)
                f = self._clone_function(f"{result.name} ({topology_mark.name})", result)
                f.stash[MarkStashKey] = topology_mark
                new_result.append(f)

            if not has_marks:
                result.stash[MarkStashKey] = None
                new_result.append(result)

        report.result = new_result

    @pytest.hookimpl(tryfirst=True)
    def pytest_collection_modifyitems(self, config: pytest.Config, items: list[pytest.Item]) -> None:
        """
        Filter collected items and deselect these that can not be run on the
        selected multihost configuration.

        Internal plugin's data are stored in ``multihost`` property of each
        :class:`pytest.Item`.

        :meta private:
        """

        selected = []
        deselected = []

        for item in items:
            data = MultihostItemData(self.multihost, item.stash[MarkStashKey]) if self.multihost else None
            MultihostItemData.SetData(item, data)

            if not self._can_run_test(item, data):
                deselected.append(item)
                continue

            selected.append(item)

        config.hook.pytest_deselected(items=deselected)
        items[:] = selected

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_setup(self, item: pytest.Item) -> None:
        """
        Create fixtures requested in :class:`~pytest_mh.TopologyMark`
        (``@pytest.mark.topology``). It adds the fixture names into ``funcargs``
        property of the pytest item in order to make them available.

        At this step, the fixtures do not have any value. The value is assigned
        later in :func:`pytest_runtest_call` hook.

        :meta private:
        """
        if not isinstance(item, pytest.Function):
            return

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        if data is None:
            return

        # Fill in parameters that will be set later in pytest_runtest_call hook,
        # otherwise pytest will raise unknown fixture error.
        if data.topology_mark is not None:
            # Make mh fixture always available
            if "mh" not in item.fixturenames:
                item.fixturenames.append("mh")

            spec = inspect.getfullargspec(item.obj)
            for arg in data.topology_mark.args:
                if arg in spec.args:
                    item.funcargs[arg] = None

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_call(self, item: pytest.Item) -> None:
        """
        Assign values to dynamically created multihost fixtures.

        :meta private:
        """
        if not isinstance(item, pytest.Function):
            return

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        if data is None:
            return

        if data.topology_mark is not None:
            mh = item.funcargs["mh"]
            if not isinstance(mh, MultihostFixture):
                raise ValueError(f"Fixture mh is not MultihostFixture but {type(mh)}!")

            data.topology_mark.apply(mh, item.funcargs)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(
        self, item: pytest.Item, call: pytest.CallInfo[None]
    ) -> Generator[None, pytest.TestReport, None]:
        """
        Store test outcome in multihost data: item.multihost.outcome. The outcome
        can be 'passed', 'failed' or 'skipped'.
        """
        outcome = yield

        data: MultihostItemData | None = MultihostItemData.GetData(item)
        result: pytest.TestReport = outcome.get_result()

        if data is None:
            raise RuntimeError("MultihostItemData should not be None")

        if result.when != "call":
            return

        data.outcome = result.outcome

    # Hook from pytest-output plugin
    @pytest.hookimpl(optionalhook=True)
    def pytest_output_item_collected(self, config: pytest.Config, item) -> None:
        try:
            from pytest_output.output import OutputDataItem
        except ImportError:
            pass

        if not isinstance(item, OutputDataItem):
            raise ValueError(f"Unexpected item type: {type(item)}")

        data: MultihostItemData | None = MultihostItemData.GetData(item.item)
        if data is None or data.topology_mark is None:
            return

        item.extra.setdefault("pytest-mh", {})["Topology"] = data.topology_mark.name

    def _fmt_color(self, text: str, color: str) -> str:
        if sys.stdout.isatty():
            reset = "\033[0m"
            return f"{color}{text}{reset}"

        return text

    def _fmt_bold(self, text: str) -> str:
        return self._fmt_color(text, "\033[1m")

    def _create_logger(self, verbose) -> logging.Logger:
        stdout = logging.StreamHandler(sys.stdout)
        stdout.setLevel(logging.DEBUG)
        stdout.setFormatter(logging.Formatter("%(message)s"))

        logger = self.GetLogger()
        logger.addHandler(stdout)
        logger.setLevel(logging.DEBUG if verbose else logging.INFO)

        return logger

    def _is_multihost_required(self, item: pytest.Item) -> bool:
        return item.get_closest_marker(name="topology") is not None

    def _can_run_test(self, item: pytest.Item, data: MultihostItemData | None) -> bool:
        if data is None:
            return not self._is_multihost_required(item)

        if self.topology is None:
            raise ValueError("Topology must be already set!")

        if data.topology_mark is not None:
            if self.mh_exact_topology:
                if data.topology_mark.topology != self.topology:
                    return False
            else:
                if not self.topology.satisfies(data.topology_mark.topology):
                    return False

        if self.mh_not_topology:
            if data.topology_mark is not None and data.topology_mark.name in self.mh_not_topology:
                return False

        if self.mh_topology:
            if data.topology_mark is None:
                return False

            if data.topology_mark.name not in self.mh_topology:
                return False

        return True

    def _clone_function(self, name: str, f: pytest.Function) -> pytest.Function:
        callspec = f.callspec if hasattr(f, "callspec") else None

        return pytest.Function.from_parent(
            parent=f.parent,
            name=name,
            callspec=callspec,
            callobj=f.obj,
            keywords=f.keywords,
            fixtureinfo=f._fixtureinfo,
            originalname=f.originalname,
        )


# These pytest hooks must be available outside of the plugin's class because
# they are executed before the plugin is registered.


def pytest_addoption(parser):
    """
    Pytest hook: add command line options.
    """

    parser.addoption("--mh-config", action="store", help="Path to the multihost configuration file")

    parser.addoption("--mh-log-path", action="store", help="Path to store multihost logs")

    parser.addoption("--mh-lazy-ssh", action="store_true", help="Postpone connecting to host SSH until it is required")

    parser.addoption(
        "--mh-topology",
        action="append",
        default=[],
        help="Filter tests by given topology, can be set multiple times",
    )

    parser.addoption(
        "--mh-not-topology",
        action="append",
        default=[],
        help="Do not run tests for given topology, can be set multiple times",
    )

    parser.addoption(
        "--mh-exact-topology",
        action="store_true",
        help="Test will be deselected if its topology does not match given multihost config exactly",
    )

    parser.addoption(
        "--mh-collect-artifacts",
        action="store",
        default="on-failure",
        nargs="?",
        choices=["never", "on-failure", "always"],
        help="Collect artifacts after test run (default: %(default)s)",
    )

    parser.addoption(
        "--mh-artifacts-dir",
        action="store",
        default="./artifacts",
        help="Directory where artifacts will be stored (default: %(default)s)",
    )


def pytest_configure(config: pytest.Config):
    """
    Pytest hook: register multihost plugin.
    """

    # register additional markers
    config.addinivalue_line(
        "markers",
        "topology(name: str, topology: pytest_mh.Topology, domains: dict[str, str], /, "
        + "*, fixture1=target1, ...): topology required to run the test",
    )

    config.addinivalue_line(
        "markers",
        "require(condition, reason): evaluate condition, parameters may be topology fixture, "
        "the test is skipped if condition is not met",
    )

    config.pluginmanager.register(MultihostPlugin(config), "MultihostPlugin")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Literal, TypeAlias

from .. import MultihostHost, MultihostRole, MultihostUtility
from ..ssh import SSHLog

__all__ = ["Firewalld"]


HostSpec: TypeAlias = str | MultihostHost | MultihostRole
"""Firewall host/hostname specification."""


ProtocolSpec: TypeAlias = Literal["tcp", "udp"]
"""Firewall protocol specification."""


PortSpec: TypeAlias = int | tuple[int, ProtocolSpec]
"""Firewall port specification."""


class Firewall(ABC, MultihostUtility):
    """
    Configure host firewall.

    All changes are automatically reverted when a test is finished.
    """

    @property
    @abstractmethod
    def inbound(self) -> FirewallInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        pass

    @property
    @abstractmethod
    def outbound(self) -> FirewallOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        pass

    def parse_port_spec(self, spec: PortSpec) -> tuple[int, ProtocolSpec]:
        """
        Parse port specification into (port, protocol) tuple.

        :raises TypeError: When invalid type is given.
        :return: (port, protocol) tuple)
        :rtype: tuple[int, Literal["tcp", "udp"]]
        """
        port: int = 0
        protocol: ProtocolSpec = "tcp"

        if isinstance(spec, int):
            port = spec
            protocol = "tcp"
        elif isinstance(spec, tuple):
            if list(map(type, spec)) == [int, str]:
                port = spec[0]
                protocol = spec[1]
            else:
                raise TypeError(f"Unexpected type: tuple{list(map(type, spec))}")
        else:
            raise TypeError(f"Unexpected type: {type(spec)}")

        if protocol not in ["tcp", "udp"]:
            raise ValueError(f"Unexpected protocol: {protocol}")

        return (port, protocol)

    def parse_host_spec(self, host: HostSpec) -> str:
        """
        Parse host specification into hostname.

        :raises ValueError: When invalid type is given.
        :return: Hostname.
        :rtype: str
        """
        if isinstance(host, str):
            return host
        elif isinstance(host, MultihostHost):
            return host.hostname
        elif isinstance(host, MultihostRole):
            return host.host.hostname
        else:
            raise ValueError(f"Unexpected type of host: {type(host)}")


class FirewallInboundRules(object):
    """
    Configure firewall inbound rules.

    All changes are automatically reverted when a test is finished.
    """

    @abstractmethod
    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept incoming traffic on given port.

        .. code-block:: python

            firewall.inbound.accept_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.accept_port((389, "tcp"))
            firewall.inbound.accept_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop incoming traffic on given port.

        .. code-block:: python

            firewall.inbound.drop_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.drop_port((389, "tcp"))
            firewall.inbound.drop_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass


class FirewallOutboundRules(object):
    """
    Configure firewall outbound rules.

    All changes are automatically reverted when a test is finished.
    """

    @abstractmethod
    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept outgoing traffic on given port.

        .. code-block:: python

            firewall.outbound.accept_port(389)  # missing protocol defaults to "tcp"
            firewall.outbound.accept_port((389, "tcp"))
            firewall.outbound.accept_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop outgoing traffic on given port.

        .. code-block:: python

            firewall.outbound.drop_port(389)  # missing protocol defaults to "tcp"
            firewall.outbound.drop_port((389, "tcp"))
            firewall.outbound.drop_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Accept outgoing traffic to given host.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an accept rule is created for each available address.

        .. code-block:: python

            firewall.outbound.accept_host("example.com")
            firewall.outbound.accept_host(["example1.com", "example2.com"])
            firewall.outbound.accept_host(multihost_host)
            firewall.outbound.accept_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        pass

    @abstractmethod
    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Drop outgoing traffic to given host.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an drop rule is created for each available address.

        .. code-block:: python

            firewall.outbound.drop_host("example.com")
            firewall.outbound.drop_host(["example1.com", "example2.com"])
            firewall.outbound.drop_host(multihost_host)
            firewall.outbound.drop_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        pass


class Firewalld(Firewall):
    """
    Configure firewall using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)

        self.__inbound: FirewalldInboundRules = FirewalldInboundRules(self)
        self.__outbound: FirewalldOutboundRules = FirewalldOutboundRules(self)

        self._priority: int = 30000
        """
        Next priority for new firewall rules, when the priority is
        auto-assigned.

        Given the nature of testing, we start with high priority and then
        decrement the value so rules can be easily overwritten (e.g. accept with
        port). E.g. adding a rule "accept port 80" and then "drop port 80" does
        not remove "accept" rule but the "drop" rule takes precedence.
        """

    def teardown_when_used(self) -> None:
        """
        Revert all firewall changes.

        :meta private:
        """
        self.host.ssh.exec(["firewall-cmd", "--reload"])
        super().teardown_when_used()

    @property
    def inbound(self) -> FirewalldInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        return self.__inbound

    @property
    def outbound(self) -> FirewalldOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        return self.__outbound

    @property
    def _next_priority(self) -> int:
        """
        Next auto-assigned priority.

        :raises ValueError: _description_
        :return: _description_
        :rtype: int
        """
        priority = self._priority
        if priority < 0:
            raise ValueError("Too many rules defined")

        self._priority -= 1
        return priority

    def add_direct_rule(
        self,
        chain: str,
        args: list[Any],
        *,
        table: str = "filter",
        ip_family: Literal["ipv4", "ipv6", "all"] = "all",
        priority: int | None = None,
    ) -> int:
        """
        Add a new direct rule.

        This methods returns a priority of this rule. You need to use this
        priority if you remove the rule with :meth:`remove_direct_rule`.

        :param chain: iptables chain (e.g. INPUT or OUTPUT).
        :type chain: str
        :param args: iptables arguments
        :type args: list[Any]
        :param table: iptables table, defaults to "filter"
        :type table: str, optional
        :param ip_family: If the rules is added as IPv4, IPv6 rule or both, defaults to ``all``.
        :type ip_family: Literal["ipv4", "ipv6", "all"], optional
        :param priority: Rule priority, defaults to None (= auto-assign next value)
        :type priority: int | None, optional
        :return: Rule priority, to be used for rule removal.
        :rtype: int
        """
        if priority is None:
            priority = self._next_priority

        cmd = [table, chain, priority, *args]

        if ip_family in ["ipv4", "all"]:
            self.logger.info(f'Firewalld: adding IPv4 direct firewall rule: {" ".join([str(x) for x in cmd])}')
            self.host.ssh.exec(["firewall-cmd", "--direct", "--add-rule", "ipv4", *cmd], log_level=SSHLog.Error)

        if ip_family in ["ipv6", "all"]:
            self.logger.info(f'Firewalld: adding IPv6 direct firewall rule: {" ".join([str(x) for x in cmd])}')
            self.host.ssh.exec(["firewall-cmd", "--direct", "--add-rule", "ipv6", *cmd], log_level=SSHLog.Error)

        return priority

    def remove_direct_rule(
        self,
        priority: int,
        chain: str,
        args: list[Any],
        *,
        table: str = "filter",
        ip_family: Literal["ipv4", "ipv6", "all"],
    ) -> None:
        """
        Remove direct rule.

        :param priority: Rule priority.
        :type priority: int
        :param chain: iptables chain (e.g. INPUT or OUTPUT).
        :type chain: str
        :param args: iptables arguments
        :type args: list[Any]
        :param table: iptables table, defaults to "filter"
        :type table: str, optional
        :param ip_family: If the rules is removed from IPv4, IPv6 rules or both, defaults to ``all``.
        :type ip_family: Literal["ipv4", "ipv6", "all"], optional
        """
        cmd = [table, chain, priority, *args]

        if ip_family in ["ipv4", "all"]:
            self.logger.info(f'Firewalld: removing IPv4 direct firewall rule: {" ".join([str(x) for x in cmd])}')
            self.host.ssh.exec(["firewall-cmd", "--direct", "--remove-rule", "ipv4", *cmd], log_level=SSHLog.Error)

        if ip_family in ["ipv6", "all"]:
            self.logger.info(f'Firewalld: removing IPv6 direct firewall rule: {" ".join([str(x) for x in cmd])}')
            self.host.ssh.exec(["firewall-cmd", "--direct", "--remove-rule", "ipv6", *cmd], log_level=SSHLog.Error)

    def add_rich_rule(self, rule: str, priority: int | None = None) -> int:
        """
        Add rich rule.

        The parameter "rule" is the part after "rule priority=X". This part is
        added automatically. That is:

        .. code-block:: console

            $ firewall-cmd --add-rich-rule rule priority={priority} {rule}

        :param rule: Firewalld rich rule.
        :type rule: str
        :param priority: Rule priority, defaults to None (= auto-assign next
            value)
        :type priority: int | None, optional
        :return: Rule priority, to be used for rule removal.
        :rtype: int
        """
        if priority is None:
            priority = self._next_priority

        rule = f"rule priority={priority} {rule}"
        self.logger.info(f'Firewalld: adding rich rule "{rule}"')
        self.host.ssh.exec(["firewall-cmd", "--add-rich-rule", rule], log_level=SSHLog.Error)

        return priority

    def remove_rich_rule(self, priority: int, rule: str) -> None:
        """
        Remove rich rule.

        The parameter "rule" is the part after "rule priority=X". This part is
        added automatically. That is:

        .. code-block:: console

            $ firewall-cmd --remove-rich-rule rule priority="{priority}" {rule}

        :param priority: Rule priority
        :type priority: int
        :param rule: Firewalld rich rule.
        :type rule: str
        """
        rule = f"rule priority={priority} {rule}"
        self.logger.info(f'Firewalld: removing rich rule  "{rule}"')
        self.host.ssh.exec(["firewall-cmd", "--remove-rich-rule", rule], log_level=SSHLog.Error)


class FirewalldInboundRules(FirewallInboundRules):
    """
    Configure firewall inbound rules using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: Firewalld) -> None:
        """
        :param firewall: Firewalld controller.
        :type firewall: Firewalld
        """
        super().__init__()

        self.firewall: Firewalld = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="accept")

    def reject_port(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Reject incoming traffic on given port.

        This blocks the communication and also sends icmp-host-unreachable.

        .. code-block:: python

            firewall.inbound.reject_port(389)  # missing protocol defaults to "tcp"
            firewall.inbound.reject_port((389, "tcp"))
            firewall.inbound.reject_port([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        self.__add_port(port, action="reject")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="drop")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["accept", "reject", "drop"],
    ) -> None:
        def __add_port(port: int, protocol: str):
            self.firewall.add_rich_rule(f'port port="{port}" protocol="{protocol}" {action}')

        def __add_service(name: str):
            self.firewall.add_rich_rule(f'service name="{name}" {action}')

        items = port if isinstance(port, list) else [port]
        for item in items:
            if isinstance(item, str):
                __add_service(item)
                continue

            port, protocol = self.firewall.parse_port_spec(item)
            __add_port(port, protocol)


class FirewalldOutboundRules(FirewallOutboundRules):
    """
    Configure firewall outbound rules using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: Firewalld) -> None:
        """
        :param firewall: Firewalld controller.
        :type firewall: Firewalld
        """
        super().__init__()

        self.firewall: Firewalld = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="ACCEPT")

    def reject_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="REJECT")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="DROP")

    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="ACCEPT")

    def reject_host(self, host: HostSpec | list[HostSpec]) -> None:
        """
        Reject outgoing traffic to given host.

        This blocks the communication and also sends icmp-host-unreachable.

        The hostname is resolved to IPv4 and IPv6 addresses via DNS and
        an drop rule is created for each available address.

        .. code-block:: python

            firewall.outbound.drop_host("example.com")
            firewall.outbound.drop_host(["example1.com", "example2.com"])
            firewall.outbound.drop_host(multihost_host)
            firewall.outbound.drop_host(multihost_role)

        :param host: Hostname, MultihostHost or MultihostRole object.
        :type host: HostSpec | list[HostSpec]
        """
        self.__add_host(host, action="REJECT")

    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="DROP")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["ACCEPT", "REJECT", "DROP"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            if isinstance(item, str):
                port = int(item)
                protocol = "tcp"
            else:
                port, protocol = self.firewall.parse_port_spec(item)

            self.firewall.add_direct_rule(
                chain="OUTPUT", args=["-p", protocol, "--dport", port, "-j", action], table="filter"
            )

    def __add_host(
        self,
        host: HostSpec | list[HostSpec],
        *,
        action: Literal["ACCEPT", "REJECT", "DROP"],
    ) -> None:
        items = host if isinstance(host, list) else [host]
        for item in items:
            hostname = self.firewall.parse_host_spec(item)
            ipv4s = self.__resolve_hostname(hostname, "A")
            ipv6s = self.__resolve_hostname(hostname, "AAAA")

            self.firewall.logger.info(
                f"Firewalld: adding {action} firewall rule for host {hostname}",
                extra={
                    "data": {
                        "Found IPv4 addresses": ipv4s,
                        "Found IPv6 addresses": ipv6s,
                    }
                },
            )

            for ip in ipv4s:
                self.firewall.add_direct_rule(
                    chain="OUTPUT", args=["--destination", ip, "-j", action], table="filter", ip_family="ipv4"
                )

            for ip in ipv6s:
                self.firewall.add_direct_rule(
                    chain="OUTPUT", args=["--destination", ip, "-j", action], table="filter", ip_family="ipv6"
                )

    def __resolve_hostname(self, hostname: str, type: Literal["A", "AAAA"]) -> list[str]:
        result = self.firewall.host.ssh.exec(["dig", "+short", "-t", type, hostname], log_level=SSHLog.Error)

        return result.stdout_lines


class WindowsFirewall(Firewall):
    """
    Configure Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)

        self.__inbound: WindowsFirewallInboundRules = WindowsFirewallInboundRules(self)
        """Manage inbound rules."""

        self.__outbound: WindowsFirewallOutboundRules = WindowsFirewallOutboundRules(self)
        """Manage outbound rules."""

        self._changed: bool = False
        """Did we change anything?"""

        self._rules: list[str] = []
        self._backup: str = "C:\\.mh_firewall.bak.wfw"

    def setup_when_used(self):
        """
        Create a backup of current firewall configuration.

        :meta private:
        """
        super().setup_when_used()
        self.logger.info(f"Windows Firewall: creating backup at '{self._backup}'")
        self.host.ssh.run(
            f"Remove-Item {self._backup}; netsh advfirewall export {self._backup}", log_level=SSHLog.Error
        )

    def teardown_when_used(self):
        """
        Revert all firewall changes.

        :meta private:
        """
        self.logger.info(f"Windows Firewall: restoring from '{self._backup}'")
        self.host.ssh.run(f"netsh advfirewall reset; netsh advfirewall import {self._backup}", log_level=SSHLog.Error)
        super().teardown_when_used()

    @property
    def inbound(self) -> WindowsFirewallInboundRules:
        """
        Configure firewall inbound rules.

        :return: Inbound rules manager.
        :rtype: FirewallInboundRules
        """
        return self.__inbound

    @property
    def outbound(self) -> WindowsFirewallOutboundRules:
        """
        Configure firewall outbound rules.

        :return: Outbound rules manager.
        :rtype: FirewalldOutboundRules
        """
        return self.__outbound

    def add_rule(
        self,
        name: str,
        direction: Literal["inbound", "outbound"],
        action: Literal["allow", "block"],
        args: list[Any],
    ) -> str:
        """
        Add firewall rule.

        Final rule name is constructed as ``"mh/{direction}/block/{name}"``.

        :param name: Rule name.
        :type name: str
        :param direction: Direction
        :type direction: Literal["inbound", "outbound"]
        :param action: Action.
        :type action: Literal["allow", "block"]
        :param args: Additional arguments to New-NetFirewallRule command.
        :type args: list[Any]
        :raises ValueError: If invalid action is given.
        :return: Final rule name.
        :rtype: str
        """
        fullname = f"mh/{direction}/{action}/{name}"
        cmd = ["New-NetFirewallRule", "-DisplayName", fullname, "-Direction", direction, "-Action", action, *args]

        match action:
            case "allow":
                opposite = f"mh/{direction}/block/{name}"
            case "block":
                opposite = f"mh/{direction}/allow/{name}"
            case _:
                raise ValueError(f"Unknown action: {action}")

        if opposite in self._rules:
            self.remove_rule(opposite)

        self.logger.info(f'Windows Firewall: adding rule: {" ".join([str(x) for x in cmd])}')
        self.host.ssh.exec(cmd, log_level=SSHLog.Error)
        self._rules.append(fullname)

        return fullname

    def remove_rule(self, name: str) -> None:
        """
        Remove firewall rule.

        :param name: Complete rule name (e.g. ``mh/inbound/block/389``)
        :type name: str
        """
        cmd = ["Remove-NetFirewallRule", "-DisplayName", name]
        self.logger.info(f'Windows Firewall: removing rule: {" ".join([str(x) for x in cmd])}')
        self.host.ssh.exec(cmd, log_level=SSHLog.Error)
        self._rules.remove(name)


class WindowsFirewallInboundRules(FirewallInboundRules):
    """
    Configure firewall inbound rules using Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: WindowsFirewall) -> None:
        """
        :param firewall: WindowsFirewall controller.
        :type firewall: WindowsFirewall
        """
        super().__init__()

        self.firewall: WindowsFirewall = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="allow")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="block")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            port, protocol = self.firewall.parse_port_spec(item)
            self.firewall.add_rule(
                f"{port}/{protocol}", "inbound", action, ["-Protocol", protocol, "-LocalPort", port]
            )


class WindowsFirewallOutboundRules(FirewallOutboundRules):
    """
    Configure firewall outbound rules using Windows Firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, firewall: WindowsFirewall) -> None:
        """
        :param firewall: WindowsFirewall controller.
        :type firewall: WindowsFirewall
        """
        super().__init__()

        self.firewall: WindowsFirewall = firewall

    def accept_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="allow")

    def drop_port(self, port: PortSpec | list[PortSpec]) -> None:
        self.__add_port(port, action="block")

    def accept_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="allow")

    def drop_host(self, host: HostSpec | list[HostSpec]) -> None:
        self.__add_host(host, action="block")

    def __add_port(
        self,
        port: PortSpec | list[PortSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = port if isinstance(port, list) else [port]
        for item in items:
            port, protocol = self.firewall.parse_port_spec(item)
            self.firewall.add_rule(
                f"{port}/{protocol}", "outbound", action, ["-Protocol", protocol, "-RemotePort", port]
            )

    def __add_host(
        self,
        host: HostSpec | list[HostSpec],
        *,
        action: Literal["allow", "block"],
    ) -> None:
        items = host if isinstance(host, list) else [host]
        for item in items:
            hostname = self.firewall.parse_host_spec(item)
            ipv4s = self.__resolve_hostname(hostname, "A")
            ipv6s = self.__resolve_hostname(hostname, "AAAA")

            self.firewall.logger.info(
                f"Windows Firewall: adding {action} firewall rule for host {hostname}",
                extra={
                    "data": {
                        "Found IPv4 addresses": ipv4s,
                        "Found IPv6 addresses": ipv6s,
                    }
                },
            )

            self.firewall.add_rule(f"{hostname}", "outbound", action, ["-RemoteAddress", ",".join([*ipv4s, *ipv6s])])

    def __resolve_hostname(self, hostname: str, type: Literal["A", "AAAA"]) -> list[str]:
        result = self.firewall.host.ssh.run(
            f"(Resolve-DnsName -Type {type} -Name {hostname}).IpAddress", log_level=SSHLog.Error
        )
        return result.stdout_lines



from __future__ import annotations

import base64
import textwrap

from .. import MultihostHost, MultihostUtility
from ..ssh import SSHLog, SSHProcessResult

__all__ = ["LinuxFileSystem"]


class LinuxFileSystem(MultihostUtility):
    """
    Perform file system operations on remote host.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)
        self.__rollback: list[str] = []
        self.__backup: dict[str, str] = {}

    def teardown(self):
        """
        Revert all file system changes.

        :meta private:
        """
        cmd = "\n".join(reversed(self.__rollback))
        if cmd:
            self.host.ssh.run(cmd)

        super().teardown()

    def mkdir(self, path: str, *, mode: str | None = None, user: str | None = None, group: str | None = None) -> None:
        """
        Create directory on remote host.

        :param path: Path of the directory.
        :type path: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        """
        self.backup(path)
        self.logger.info(f'Creating directory "{path}"')
        self.host.ssh.run(
            f"""
                set -ex
                rm -fr '{path}'
                mkdir '{path}'
                {self.__gen_chattrs(path, mode=mode, user=user, group=group)}
            """,
            log_level=SSHLog.Error,
        )

    def mkdir_p(
        self, path: str, *, mode: str | None = None, user: str | None = None, group: str | None = None
    ) -> None:
        """
        Create directory on remote host, including all missing parent directories.

        :param path: Path of the directory.
        :type path: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        """
        self.backup(path)
        self.logger.info(f'Creating directory "{path}" (with parents)')
        result = self.host.ssh.run(
            f"""
                set -ex
                rm -fr '{path}'
                mkdir -v -p '{path}' | head -1 | sed -E "s/mkdir:[^']+'(.+)'$/\\1/"
                {self.__gen_chattrs(path, mode=mode, user=user, group=group)}
            """,
            log_level=SSHLog.Error,
        )

        if result.stdout:
            self.__rollback.append(f"rm --force --recursive '{result.stdout}'")

    def mktmp(
        self,
        contents: str | None = None,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
        dedent: bool = True,
    ) -> str:
        """
        Create temporary file on remote host.

        :param contents: File contents to write.
        :type contents: str | None
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        :param dedent: Automatically dedent and strip file contents, defaults to True
        :type dedent: bool, optional
        :raises OSError: If the file can not be created.
        :return: Temporary file path.
        :rtype: str
        """

        self.logger.info("Creating temporary file")
        result = self.host.ssh.run(
            """
                set -ex
                tmp=`mktemp /tmp/mh.fs.rollback.XXXXXXXXX`
                echo $tmp
            """,
            log_level=SSHLog.Error,
        )

        tmpfile = result.stdout.strip()
        if not tmpfile:
            raise OSError("Temporary file was not created")

        self.__rollback.append(f"rm --force '{tmpfile}'")

        if contents is not None:
            if dedent:
                contents = textwrap.dedent(contents).strip()

            self.logger.info(f'Writing file "{tmpfile}"', extra={"data": {"Contents": contents}})
            self.host.ssh.run(f"cat > '{tmpfile}'", input=contents, log_level=SSHLog.Error)

        attrs = self.__gen_chattrs(tmpfile, mode=mode, user=user, group=group)
        if attrs:
            self.host.ssh.run(attrs, log_level=SSHLog.Error)

        return tmpfile

    def rm(self, path: str) -> None:
        """
        Remove remote file or directory.

        :param path: File path.
        :type path: str
        """
        self.backup(path)
        self.logger.info(f'Removing file "{path}"')

        self.host.ssh.run(
            f"""
                set -ex
                rm -fr '{path}'
            """,
            log_level=SSHLog.Error,
        )

    def read(self, path: str) -> str:
        """
        Read remote file and return its contents.

        :param path: File path.
        :type path: str
        :return: File contents.
        :rtype: str
        """
        self.logger.info(f'Reading file "{path}"')
        result = self.host.ssh.exec(["cat", path], log_level=SSHLog.Error)

        return result.stdout

    def exists(self, path: str) -> bool:
        """
        Checks file or directory to see if they exist.

        :param path: File path.
        :type path: str
        :return: True or False
        :rtype: bool
        """
        self.logger.info(f'Checking if "{path}" exists')
        result = self.host.ssh.exec(["ls", path], log_level=SSHLog.Error, raise_on_error=False)

        if result.rc == 0:
            return True

        return False

    def write(
        self,
        path: str,
        contents: str,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
        dedent: bool = True,
    ) -> None:
        """
        Write to a remote file.

        :param path: File path.
        :type path: str
        :param contents: File contents to write.
        :type contents: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        :param dedent: Automatically dedent and strip file contents, defaults to True
        :type dedent: bool, optional
        """
        if dedent:
            contents = textwrap.dedent(contents).strip()

        self.backup(path)
        self.logger.info(f'Writing file "{path}"', extra={"data": {"Contents": contents}})

        self.host.ssh.run(
            f"""
                set -ex
                rm -fr '{path}'
                cat > '{path}'
                {self.__gen_chattrs(path, mode=mode, user=user, group=group)}
            """,
            input=contents,
            log_level=SSHLog.Error,
        )

    def append(
        self,
        path: str,
        contents: str,
        *,
        dedent: bool = True,
    ) -> None:
        """
        Append to a remote file.

        :param path: File path.
        :type path: str
        :param contents: File contents to write.
        :type contents: str
        :param dedent: Automatically dedent and strip file contents, defaults to True
        :type dedent: bool, optional
        """
        if dedent:
            contents = textwrap.dedent(contents).strip()

        self.backup(path)
        self.logger.info(f'Appending to file "{path}"', extra={"data": {"Contents": contents}})

        self.host.ssh.run(
            f"""
                set -ex
                cat >> '{path}'
            """,
            input=contents,
            log_level=SSHLog.Error,
        )

    def touch(
        self,
        path: str,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
    ) -> None:
        """
        Touch a remote file.

        :param path: File path.
        :type path: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        :param dedent: Automatically dedent and strip file contents, defaults to True
        :type dedent: bool, optional
        """
        self.backup(path)
        self.logger.info(f'Touching file "{path}"')

        self.host.ssh.run(
            f"""
                set -ex
                touch '{path}'
                {self.__gen_chattrs(path, mode=mode, user=user, group=group)}
            """,
            log_level=SSHLog.Error,
        )

    def truncate(
        self,
        path: str,
        *,
        size: int = 0,
    ) -> None:
        """
        Truncate remote file.

        :param path: File path.
        :type path: str
        :param size: Target file size, defaults to 0
        :type size: int, optional
        """
        self.backup(path)
        self.logger.info(f'Truncating file "{path}"', extra={"data": {"Size": size}})

        self.host.ssh.run(
            f"""
                set -ex
                truncate -s '{size}' '{path}'
            """,
            log_level=SSHLog.Error,
        )

    def copy(
        self,
        srcpath: str,
        dstpath: str,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
    ) -> None:
        """
        Copy a remote file @srcpath to remote @dstpath.

        :param srcpath: Remote source file path.
        :type srcpath: str
        :param dstpath: Remote destination file path.
        :type dstpath: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        :param dedent: Automatically dedent and strip file contents, defaults to True
        :type dedent: bool, optional
        """
        self.backup(dstpath)
        self.logger.info(f'Copying file "{srcpath}" to "{dstpath}"')

        self.host.ssh.run(
            f"""
                set -ex
                cp --archive '{srcpath}' '{dstpath}'
                {self.__gen_chattrs(dstpath, mode=mode, user=user, group=group)}
            """,
            log_level=SSHLog.Error,
        )

    def upload(
        self,
        local_path: str,
        remote_path: str,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
    ) -> None:
        """
        Upload local file.

        :param local_path: Source local path.
        :type local_path: str
        :param remote_path: Destination remote path.
        :type remote_path: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        """
        self.backup(remote_path)
        self.logger.info(f'Uploading file "{local_path}" to "{self.host.hostname}:{remote_path}"')
        with open(local_path, "rb") as f:
            encoded = base64.b64encode(f.read()).decode("utf-8")

        self.host.ssh.run(
            f"""
                set -ex
                rm -fr '{remote_path}'
                base64 --decode > '{remote_path}'
                {self.__gen_chattrs(remote_path, mode=mode, user=user, group=group)}
            """,
            input=encoded,
            log_level=SSHLog.Error,
        )
        self.__rollback.append(f"rm --force '{remote_path}'")

    def upload_to_tmp(
        self,
        local_path: str,
        *,
        mode: str | None = None,
        user: str | None = None,
        group: str | None = None,
    ) -> str:
        """
        Upload local file to a new temporary file on remote host.

        :param local_path: Source local path.
        :type local_path: str
        :param mode: Access mode (chmod value), defaults to None
        :type mode: str | None, optional
        :param user: Owner, defaults to None
        :type user: str | None, optional
        :param group: Group, defaults to None
        :type group: str | None, optional
        :return: Temporary file path.
        :rtype: str
        """
        tmp_path = self.mktmp(mode=mode, user=user, group=group)

        self.logger.info(f'Uploading file "{local_path}" to "{self.host.hostname}:{tmp_path}"')
        with open(local_path, "rb") as f:
            encoded = base64.b64encode(f.read()).decode("utf-8")

        self.host.ssh.run(f"base64 --decode > '{tmp_path}'", input=encoded, log_level=SSHLog.Error)

        return tmp_path

    def download(self, remote_path: str, local_path: str) -> None:
        """
        Download file from remote host to local machine.

        :param remote_path: Remote path.
        :type remote_path: str
        :param local_path: Local path.
        :type local_path: str
        """
        self.logger.info(f'Downloading file "{self.host.hostname}:{remote_path}" to "{local_path}"')
        result = self.host.ssh.exec(["base64", remote_path], log_level=SSHLog.Error)
        with open(local_path, "wb") as f:
            f.write(base64.b64decode(result.stdout))

    def download_files(self, paths: list[str], local_path: str) -> None:
        """
        Download multiple files from remote host. The files are stored in single
        gzipped tarball on the local machine. The remote file path may contain
        glob pattern.

        :param paths: List of remote file paths. May contain glob pattern.
        :type paths: list[str]
        :param local_path: Path to the gzipped tarball destination file on local machine.
        :type local_path: str
        """
        self.logger.info(
            f'Downloading files from {self.host.hostname} to "{local_path}"', extra={"data": {"Paths": paths}}
        )
        result = self.host.ssh.run(
            f"""
            tmp=`mktemp /tmp/mh.fs.download_files.XXXXXXXXX`
            tar -czvf "$tmp" {' '.join([f'$(compgen -G "{path}")' for path in paths])} &> /dev/null
            base64 "$tmp"
            rm -f "$tmp" &> /dev/null
        """,
            log_level=SSHLog.Error,
        )

        with open(local_path, "wb") as f:
            f.write(base64.b64decode(result.stdout))

    def backup(self, path: str) -> bool:
        """
        Backup file or directory.

        The path is automatically restored from the backup when a test is
        finished.

        .. note::
            It is also possible that the file or directory does not exist. In
            that case, the path is removed during the teardown process to
            remove any file or directory that might have been created.

        :param path: Path to back up.
        :type path: str
        :return: True if the path exists and backup was done, False otherwise.
        :rtype: bool
        """
        if path in self.__backup:
            # Backup is already present
            return True

        self.logger.info(f'Creating a backup of "{path}"')
        result = self.host.ssh.run(
            f"""
        set -ex

        if [ -f '{path}' ]; then
            tmp=`mktemp /tmp/mh.fs.rollback.XXXXXXXXX`
            cp --force --archive '{path}' "$tmp"
            echo "mv --force '$tmp' '{path}'"
        elif [ -d '{path}' ]; then
            tmp=`mktemp -d /tmp/mh.fs.rollback.XXXXXXXXX`
            cp --force --archive '{path}/.' "$tmp"
            echo "rm --force --recursive '{path}' && mv --force '$tmp' '{path}'"
        elif [ ! -d '{path}' ] && [ ! -f '{path}' ]; then
            echo "rm --force --recursive '{path}'"
        fi
        """,
            log_level=SSHLog.Error,
        )

        action = result.stdout.strip()
        if action:
            self.__rollback.append(action)
            self.__backup[path] = action
            return True

        return False

    def restore(self, path: str) -> bool:
        """
        Restore file or directory from previous backup.

        .. note::
            It is also possible that the file or directory does not exist. In
            that case, the path is removed to remove any file or directory that
            might have been created.

        :param path: Path to restore.
        :type path: str
        :return: True if the backup of path exists and it was restored, False otherwise.
        :rtype: bool
        """
        action = self.__backup.get(path)
        if action is None:
            # Backup is not present
            return False

        self.logger.info(f'Restoring "{path}" from backup')
        self.host.ssh.run(action, log_level=SSHLog.Error)

        self.__rollback.remove(action)
        del self.__backup[path]

        return True

    def __gen_chattrs(
        self, path: str, *, mode: str | None = None, user: str | None = None, group: str | None = None
    ) -> str:
        cmds = []
        if mode is not None:
            cmds.append(f"chmod '{mode}' '{path}'")

        if user is not None:
            cmds.append(f"chown '{user}' '{path}'")

        if group is not None:
            cmds.append(f"chgrp '{group}' '{path}'")

        return " && ".join(cmds)

    def wc(
        self, file: str, lines: bool = False, word: bool = False, bytes: bool = False, chars: bool = False
    ) -> SSHProcessResult:
        """
        Print newline, word, and byte counts for specific file.

        Output example without additional arguments: ``67 564 3514 file_name``

        :param file: File whose content is counted
        :type file: str
        :param lines: Print the newline counts, defaults to False
        :type lines: bool, optional
        :param word: Print the word counts, defaults to False
        :type word: bool, optional
        :param bytes: Print the byte counts, defaults to False
        :type bytes: bool, optional
        :param chars: Print the character counts, defaults to False
        :type chars: bool, optional
        :return: Result of process
        :rtype: SSHProcessResult
        """
        args = []
        if lines:
            args.append("-l")

        if word:
            args.append("-w")

        if bytes:
            args.append("-b")

        if chars:
            args.append("-m")

        return self.host.ssh.exec(["wc", *args, file], log_level=SSHLog.Error)

    def chmod(self, mode: str, path: str, args: list[str] | None = None) -> SSHProcessResult:
        """
        Change file/folder mode bits.
        Mode can be specified in two ways: octal number e.g. "666", "444" or
        a symbolic representation of changes e.g. "u=rw,go=r", "go-rw"

        :param mode: New mode of file/folder
        :type mode: str
        :param path: File or folder whose permissions change
        :type path: str
        :param args: Additional options, defaults to None
        :type args: list[str] | None, optional
        :return: Result of process
        :rtype: SSHProcessResult
        """
        self.backup(path)
        self.logger.info(f'Changing mode to "{mode}" for "{path}"')
        args = args if args else []
        return self.host.ssh.exec(["chmod", *args, mode, path], log_level=SSHLog.Error)

    def chown(
        self, path: str, user: str | None = None, group: str | None = None, args: list[str] | None = None
    ) -> SSHProcessResult:
        """
        Change file owner and group.

        :param path: Path to file
        :type path: str
        :param user: New file owner, if None then user remains same, defaults to None
        :type user: str | None, optional
        :param group: New file group, if None then group remains same, defaults to None
        :type group: str | None, optional
        :param args: Additional options, defaults to None
        :type args: list[str] | None, optional
        :return: Result of process
        :rtype: SSHProcessResult
        """
        self.backup(path)
        if user:
            self.logger.info(f'Changing owner of "{path}" to "{user}"')
        if group:
            self.logger.info(f'Changing group of "{path}" to "{group}"')

        args = args if args else []
        mode = f"{user}" if user else ""
        mode += f":{group}" if group else ""
        return self.host.ssh.exec(["chown", mode, *path.split(), *args], log_level=SSHLog.Error)



from __future__ import annotations

from .. import MultihostHost, MultihostUtility
from ..ssh import SSHLog

__all__ = ["LinuxTrafficControl"]


class LinuxTrafficControl(MultihostUtility):
    """
    Perform traffic control operations on remote host.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)
        self.__interfaces: set[str] = set()
        self.__restore_root: set[str] = set()
        self.__restore_filters: dict[str, int] = dict()
        self.__band: int = 2

    def setup_when_used(self) -> None:
        """
        Setup traffic control configuration.

        :meta private:
        """
        super().setup_when_used()

        # Let's find out the available interfaces
        result = self.host.ssh.run("ip -o address", log_level=SSHLog.Error)
        split_result = result.stdout.splitlines()

        for line in split_result:
            interface = line.split(" ")[1]
            if interface != "lo":
                self.__interfaces.add(interface)

        commands = "set -e\n"
        for interface in self.__interfaces:
            commands += (
                f"tc qdisc add dev {interface} root handle 1: prio bands 16 priomap 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n"
            )
            self.__restore_root.add(f"tc qdisc del dev {interface} root")

        self.host.ssh.run(commands, log_level=SSHLog.Error)

    def teardown_when_used(self) -> None:
        """
        Revert all traffic control changes.

        :meta private:
        """
        tear: str = ""
        for iter in self.__restore_root:
            tear += iter + "\n"
        self.host.ssh.run(tear, log_level=SSHLog.Error)
        super().teardown_when_used()

    def add_delay(self, hostname: str, time: str | int):
        """
        Add delay to the network connection. A maximum of 15 connections can be delayed at a time. It is recommended
        to specify the delay from minimum to maximum to avoid starvation.

        :param hostname: Target hostname.
        :type hostname: str
        :param time: Delay. Units can be specified; if not specified, the default is milliseconds.
        :type time: str | int
        """
        if isinstance(time, int):
            time_unit = f"{time}ms"
        else:
            time_unit = time

        self.logger.info(f"Adding network delay {time_unit} to {hostname}")

        ips = self.host.ssh.run(f"dig +short {hostname}", log_level=SSHLog.Error)
        ip_list = ips.stdout.splitlines()

        commands = "set -e\n"
        for interface in self.__interfaces:
            commands += (
                f"tc qdisc add dev {interface} parent 1:{self.__band} handle {self.__band*10}: netem "
                f"delay {time_unit}\n"
            )
            for ip in ip_list:
                commands += (
                    f"tc filter add dev {interface} protocol ip prio {self.__band} parent 1:0 u32 match ip dst {ip} "
                    f"flowid 1:{self.__band}\n"
                )
                self.__restore_filters[ip] = self.__band

        self.host.ssh.run(commands, log_level=SSHLog.Error)
        self.__band += 1

    def remove_delay(self, hostname: str):
        """
        Remove delay in the network connection.

        :param hostname: Target hostname.
        :type hostname: str
        """
        self.logger.info(f"Removing network delay to {hostname}")

        ips = self.host.ssh.run(f"dig +short {hostname}", log_level=SSHLog.Error)
        ip_list = ips.stdout.splitlines()

        bands: set[int] = set()
        for ip in ip_list:
            if ip in self.__restore_filters:
                bands.add(self.__restore_filters[ip])

        commands = "set -e\n"
        for interface in self.__interfaces:
            for band in bands:
                commands += f"tc filter del dev {interface} prio {band}\n"

            commands += f"tc qdisc del dev {interface} parent 1:{band} handle {band*10}: netem\n"

        self.host.ssh.run(commands, log_level=SSHLog.Error)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Mapping, Tuple

import pytest

if TYPE_CHECKING:
    from .fixtures import MultihostFixture
    from .topology import Topology


class TopologyMark(object):
    """
    Topology mark is used to describe test case requirements. It defines:

    * **name**, that is used to identify topology in pytest output
    * **topology** (:class:Topology) that is required to run the test
    * **fixtures** that are available during the test run

    .. code-block:: python
        :caption: Example usage

        @pytest.mark.topology(name, topology, fixture1='path1', fixture2='path2', ...)
        def test_fixture_name(fixture1: BaseRole, fixture2: BaseRole, ...):
            assert True

    Fixture path points to a host in the multihost configuration and can be
    either in the form of ``$domain-id.$role`` (all host of given role) or
    ``$domain-id.$role[$index]`` (specific host on given index).

    The ``name`` is visible in verbose pytest output after the test name, for example:

    .. code-block:: console

        tests/test_basic.py::test_case (topology-name) PASSED
    """

    def __init__(
        self,
        name: str,
        topology: Topology,
        fixtures: dict[str, str] | None = None,
    ) -> None:
        """
        :param name: Topology name used in pytest output.
        :type name: str
        :param topology: Topology required to run the test.
        :type topology: Topology
        :param fixtures: Dynamically created fixtures available during the test run, defaults to None
        :type fixtures: dict[str, str] | None, optional
        """
        self.name: str = name
        """Topology name."""

        self.topology: Topology = topology
        """Multihost topology."""

        self.fixtures: dict[str, str] = fixtures if fixtures is not None else {}
        """Dynamic fixtures mapping."""

        self.mapping: dict[str, list[str]] = {}

        for fixture, target in self.fixtures.items():
            self.mapping.setdefault(target, list()).append(fixture)

    @property
    def args(self) -> set[str]:
        """
        Names of all dynamically created fixtures.
        """

        return set(self.fixtures.keys())

    def apply(self, mh: MultihostFixture, funcargs: dict[str, Any]) -> None:
        """
        Create required fixtures by modifying pytest.Item.funcargs.

        :param mh: _description_
        :type mh: Multihost
        :param funcargs: Pytest test item ``funcargs`` that will be modified.
        :type funcargs: dict[str, Any]
        """

        for path, names in self.mapping.items():
            value = mh._lookup(path)
            for name in names:
                if name in funcargs:
                    funcargs[name] = value

    def export(self) -> dict:
        """
        Export the topology mark into a dictionary object that can be easily
        converted to JSON, YAML or other formats.

        .. code-block:: python

            {
                'name': 'client',
                'fixtures': { 'client': 'test.client[0]' },
                'topology': [
                    {
                        'id': 'test',
                        'hosts': { 'client': 1 }
                    }
                ]
            }

        :rtype: dict
        """

        return {
            "name": self.name,
            "fixtures": self.fixtures,
            "topology": self.topology.export(),
        }

    @classmethod
    def ExpandMarkers(cls, item: pytest.Item) -> list[pytest.Mark]:
        out = []
        for mark in item.iter_markers("topology"):
            # We need to use generic classes in order to avoid circular import
            if isinstance(mark.args[0], KnownTopologyGroupBase) and isinstance(mark.args[0].value, list):
                for topology in mark.args[0].value:
                    out.append(pytest.mark.topology(topology))
                continue

            out.append(mark)

        return out

    @classmethod
    def Create(cls, item: pytest.Function, mark: pytest.Mark) -> TopologyMark:
        """
        Create instance of :class:`TopologyMark` from ``@pytest.mark.topology``.

        :raises ValueError:
        :rtype: TopologyMark
        """
        if not mark.args or len(mark.args) > 3:
            nodeid = item.parent.nodeid if item.parent is not None else ""
            error = f"{nodeid}::{item.originalname}: invalid arguments for @pytest.mark.topology"
            raise ValueError(error)

        # Constructor for KnownTopologyBase
        if isinstance(mark.args[0], KnownTopologyBase):
            if len(mark.args) != 1:
                raise ValueError(error)

            if not isinstance(mark.args[0].value, cls):
                raise ValueError(error)

            return mark.args[0].value

        # Generic constructor.
        return cls._CreateFromArgs(item, mark.args, mark.kwargs)

    @classmethod
    def _CreateFromArgs(cls, item: pytest.Function, args: Tuple, kwargs: Mapping[str, Any]) -> TopologyMark:
        """
        Create :class:`TopologyMark` from pytest marker arguments.

        .. warning::

            This should only be called internally. You can inherit from
            :class:`TopologyMark` and override this in order to add additional
            attributes to the marker.

        :param item: Pytest item.
        :type item: pytest.Function
        :param args: Pytest mark positional arguments.
        :type args: Any
        :param kwargs: Pytest mark keyword arguments arguments.
        :type kwargs: Mapping[str, Any]
        :raises ValueError: If the marker is invalid.
        :return: Instance of TopologyMark.
        :rtype: TopologyMark
        """
        # First three parameters are positional, the rest are keyword arguments.
        if len(args) != 2:
            nodeid = item.parent.nodeid if item.parent is not None else ""
            error = f"{nodeid}::{item.originalname}: invalid arguments for @pytest.mark.topology"
            raise ValueError(error)

        name = args[0]
        topology = args[1]
        fixtures = {k: str(v) for k, v in kwargs.items()}

        return cls(name, topology, fixtures)


class KnownTopologyBase(Enum):
    """
    Base class for a predefined set of topologies.

    Users of this plugin may inherit from this class in order to created a
    predefined, well-known set of topology markers.

    .. code-blocK:: python
        :caption: Example usage

        @final
        @unique
        class KnownTopology(KnownTopologyBase):
            A = TopologyMark(
                name='A',
                topology=Topology(TopologyDomain('test', a=1)),
                fixtures=dict(a='test.a[0]'),
            )

            B = TopologyMark(
                name='B',
                topology=Topology(TopologyDomain('test', b=1)),
                fixtures=dict(b='test.b[0]'),
            )


        @pytest.mark.topology(KnownTopology.A)
        def test_a(a: ARole):
            pass

        @pytest.mark.topology(KnownTopology.B)
        def test_b(b: BRole):
            pass
    """

    pass


class KnownTopologyGroupBase(Enum):
    """
    Base class for a predefined set of list of topologies.

    Users of this plugin may inherit from this class in order to create a
    predefined, well-known set of list of topology markers that can be used
    directly in ``@pytest.mark.topology`` to enable topology parametrization for
    a test case.

    .. code-blocK:: python
        :caption: Example usage

        @final
        @unique
        class KnownTopologyGroup(KnownTopologyGroupBase):
            All = [
                TopologyMark(
                    name='A',
                    topology=Topology(TopologyDomain('test', a=1)),
                    fixtures=dict(a='test.a[0]', generic='test.a[0]'),
                ),

                B = TopologyMark(
                    name='B',
                    topology=Topology(TopologyDomain('test', b=1)),
                    fixtures=dict(b='test.b[0]', generic='test.a[0]'),
                )
            ]


        # Will run once for A, once for B
        @pytest.mark.topology(KnownTopologyGroup.All)
        def test_all(generic: GenericRole):
            pass
    """

    pass



from __future__ import annotations

import logging
import textwrap
from logging.handlers import MemoryHandler
from pathlib import Path
from typing import Any, Type

import colorama

from .misc import merge_dict


class MultihostLogger(logging.Logger):
    """
    Multihost logger class.

    It extends the standard logger with additional :meth:`colorize` method that
    can be used to put some colors into the log message.

    It also allows to log extra data, that are printed in a formatted way
    together with the message.

    .. code-block:: python

        logger.info(
            'Main message'',
            extra={'data': {
                'Field1': 'value1',
                'Field2': 'value2',
                ...
            }}
        )
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.allow_colors: bool = False
        self.log_path: str | None = None
        self.extra: dict[str, Any] = {}
        self.handler: logging.Handler | None = None

    @classmethod
    def GetLogger(
        cls, *, loggercls: Type[MultihostLogger] | None = None, suffix: str | None = None
    ) -> MultihostLogger:
        """
        Returns the multihost logger.

        :param loggercls: Logger class, defaults to None (= cls).
        :type loggercls: Type[MultihostLogger] | None
        :param suffix: Logger name suffix, defaults to None.
        :type suffix: str | None
        :return: Logger.
        :rtype: MultihostLogger
        """
        name = "pytest_mh.logger"
        if suffix:
            name += f".{suffix}"

        if loggercls is None:
            loggercls = cls

        old_class = logging.getLoggerClass()

        logging.setLoggerClass(loggercls)
        logger = logging.getLogger(name)
        logging.setLoggerClass(old_class)

        if not isinstance(logger, loggercls):
            raise ValueError(f"logger must be instance of {loggercls.__name__}")

        return logger

    def setup(self, **kwargs) -> None:
        """
        Setup multihost logging facility.

        Colors are allowed if ``log_path`` is ``/dev/stdout`` or ``/dev/stderr``.

        :param log_path: Path to the log file.
        :type log_path: str
        """
        host_length = self._max_host_length(kwargs.get("confdict", {}))
        self.log_path = kwargs["log_path"]
        self.allow_colors = self.log_path in ["/dev/stdout", "/dev/stderr"]
        self.handler = ManualMemoryHandler() if self.log_path is None else logging.FileHandler(self.log_path)

        self.handler.setLevel(logging.DEBUG)
        self.handler.setFormatter(
            logging.Formatter(f"%(levelname)-8s %(asctime)s %(host){host_length}s %(message)s", defaults={"host": ""})
        )

        self.addHandler(self.handler)
        self.addFilter(LogExtraDataFilter(logger=self, indent=34 + host_length))
        self.setLevel(logging.DEBUG)

    def subclass(self, cls: Type[MultihostLogger], suffix: str, **kwargs) -> MultihostLogger:
        logger: MultihostLogger = self.GetLogger(loggercls=cls, suffix=suffix)

        # Copy only level and filters, we don't want subloggers to have handlers
        logger.handlers = []
        logger.filters = list(self.filters)
        logger.level = self.level

        # Copy selected attributes
        logger.allow_colors = self.allow_colors
        logger.log_path = self.log_path
        logger.extra = dict(self.extra)

        logger.setup(**kwargs)

        return logger

    def colorize(self, text: str | Any, *colors: str) -> str:
        """
        Make the ``text`` colored with ANSI colors.

        :param text: Text to format. ``str(text)`` is called on the parameter.
        :type text: str | Any
        :param \\*colors: Colors to apply on the text.
        :type \\*colors: colorama.Fore | colorama.Back | colorama.Style
        :return: Text with colors, if colors are allowed. Unchanged text otherwise.
        :rtype: str
        """
        if not self.allow_colors:
            return str(text)

        return "".join(colors) + str(text) + colorama.Style.RESET_ALL

    def debug(self, msg, *args, **kwargs):
        super().debug(msg, *args, **self._msgdata(kwargs))

    def info(self, msg, *args, **kwargs):
        super().info(msg, *args, **self._msgdata(kwargs))

    def warning(self, msg, *args, **kwargs):
        super().warning(msg, *args, **self._msgdata(kwargs))

    def warn(self, msg, *args, **kwargs):
        super().warn(msg, *args, **self._msgdata(kwargs))

    def error(self, msg, *args, **kwargs):
        super().error(msg, *args, **self._msgdata(kwargs))

    def exception(self, msg, *args, exc_info=True, **kwargs):
        super().exception(msg, *args, exc_info=exc_info, **self._msgdata(kwargs))

    def critical(self, msg, *args, **kwargs):
        super().critical(msg, *args, **self._msgdata(kwargs))

    def fatal(self, msg, *args, **kwargs):
        super().fatal(msg, *args, **self._msgdata(kwargs))

    def log(self, level, msg, *args, **kwargs):
        super().warning(level, msg, *args, **self._msgdata(kwargs))

    def clear(self) -> None:
        """
        Clear current log records buffer without writing it anywhere.
        """
        if isinstance(self.handler, ManualMemoryHandler):
            self.handler.clear()

    def write_to_file(self, path: str) -> None:
        """
        Write current log records buffer to a file and clear the buffer.

        :param path: Destination file path.
        :type path: str
        """
        if isinstance(self.handler, ManualMemoryHandler):
            self.handler.write_to_file(path)

    def _msgdata(self, kwargs) -> dict[str, Any]:
        if self.extra:
            return merge_dict(kwargs, {"extra": self.extra})

        return kwargs

    def _max_host_length(self, confict: dict) -> int:
        length = 0
        for domain in confict.get("domains", []):
            for host in domain.get("hosts", []):
                length = max(length, len(host.get("hostname", "")))

        return length


class MultihostHostLogger(MultihostLogger):
    """
    Logger for individual hosts.

    It includes hostname as ``host`` key in the message extra data.
    """

    def setup(self, **kwargs) -> None:
        self.extra["host"] = kwargs["hostname"]


class LogExtraDataFilter(logging.Filter):
    """
    :meta private:
    """

    def __init__(self, *args, logger: MultihostLogger, indent: int, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.indent: int = indent
        self.__logger: MultihostLogger = logger

    def dumps(self, o) -> str:
        if isinstance(o, dict):
            out = ""
            for key, value in o.items():
                value = str(value)
                out += "\n"
                out += textwrap.indent(f"{self.__logger.colorize(key, colorama.Fore.BLUE)}: {value}", " " * 2)

            return out

        if isinstance(o, (list, set, tuple)):
            out = ""
            for value in o:
                out += "\n- " + str(value)

            return out

        value = str(o)
        if "\n" not in value:
            return value

        return "|\n" + textwrap.indent(value, " " * 2)

    def filter(self, record):
        if hasattr(record, "data"):
            for key, value in record.data.items():
                record.msg += "\n"
                record.msg += textwrap.indent(
                    f"{self.__logger.colorize(key, colorama.Fore.MAGENTA)}: {self.dumps(value)}", " " * self.indent
                )

        return super().filter(record)


class ManualMemoryHandler(MemoryHandler):
    def __init__(self) -> None:
        super().__init__(capacity=0, flushLevel=0, target=None, flushOnClose=False)

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        """
        This handler does not flush automatically.

        :param record: Log record. Unused.
        :type record: logging.LogRecord
        :return: Always return False.
        :rtype: bool
        """
        return False

    def clear(self) -> None:
        """
        Clear current buffer without writing it anywhere.
        """
        self.buffer.clear()

    def write_to_file(self, path: str) -> None:
        """
        Write current buffer to a file and clear the buffer.

        :param path: Destination file path.
        :type path: str
        """
        Path(path).parent.mkdir(parents=True, exist_ok=True)

        handler = logging.FileHandler(path)
        handler.setLevel(self.level)
        handler.setFormatter(self.formatter)

        self.setTarget(handler)
        self.flush()
        self.setTarget(None)



from __future__ import annotations

from collections import Counter


class TopologyDomain(object):
    """
    Create a new topology domain.

    Topology domain specifies domain id required by the topology as well as
    required roles and number of hosts that must implement these roles. See
    :class:`Topology` for more information.

    The following example defines a topology domain of id ``test`` that
    requires two roles: ``client`` and ``ldap`` each provided by one host.

    .. code-block:: python

        TopologyDomain(
            'test',
            client=1, ldap=1
        )
    """

    def __init__(self, id: str, **kwargs: int) -> None:
        """
        :param id: Domain id.
        :type id: str
        :param `*kwargs`: Required roles.
        :type `*kwargs`: dict[str, int]
        """

        self.id: str = id
        self.roles: dict[str, int] = kwargs

    def get(self, role: str) -> int:
        """
        Find role and return the number of hosts that must implement this role.

        :param role: Host role to lookup.
        :type rol: str
        :raises KeyError: The domain was not found.
        :rtype: int
        """

        return self.roles[role]

    def export(self) -> dict:
        """
        Export the topology domain into a dictionary object that can be easily
        converted to JSON, YAML or other formats.

        .. code-block:: python

            {
                'id': 'test',
                'roles': {
                    'client': 1,
                    'ldap': 1
                }
            }

        :rtype: dict
        """

        return {"id": self.id, "hosts": self.roles}

    def satisfies(self, other: "TopologyDomain") -> bool:
        """
        Check if the topology domain satisfies the ``other`` domain.

        Returns ``True`` if the domain ids match and this domain contains all
        required roles defined in the ``other`` topology and ``False``
        otherwise.

        :param other: The other topology domain.
        :type other: TopologyDomain
        :rtype: bool
        """

        if self.id != other.id:
            return False

        for role, value in other.roles.items():
            if role not in self or self.get(role) < value:
                return False

        return True

    def __str__(self) -> str:
        return str(self.export())

    def __contains__(self, item: str) -> bool:
        return item in self.roles

    def __eq__(self, other) -> bool:
        return self.export() == other.export()

    def __ne__(self, other) -> bool:
        return self.export() != other.export()


class Topology(object):
    """
    A topology specifies requirements that a multihost configuration must fulfil
    in order to run a test.

    Each topology consist of one or more domains (:class:`TopologyDomain`) that
    defines how many hosts are available inside the domain and what roles are
    implemented.

    The following example defines an ldap topology that consist of one domain of
    id ``test`` and requires two roles: ``client`` and ``ldap`` each provided
    by one host.

    .. code-block:: python

        Topology(
            TopologyDomain(
                'test',
                 client=1, ldap=1
            )
        )

    This topology can be satisfied for example by the following multihost
    configuration:

    .. code-block:: yaml

        domains:
        - name: ldap.test
          id: test
          hosts:
          - name: client
            hostname: client.test
            role: client

          - name: ldap
            hostname: master.ldap.test
            role: ldap
    """

    def __init__(self, *domains: TopologyDomain) -> None:
        """
        :param `*args`: Domains that are included in this topology.
        :type `*args`: TopologyDomain
        """

        self.domains = list(domains)

    def get(self, id: str) -> TopologyDomain:
        """
        Find topology domain of the given id and return it.

        :param id: Topology domain id to lookup.
        :type id: str
        :raises KeyError: The domain was not found.
        :rtype: TopologyDomain
        """

        for domain in self.domains:
            if domain.id == id:
                return domain

        raise KeyError(f'Domain "{id}" was not found.')

    def export(self) -> list[dict]:
        """
        Export the topology into a list of dictionaries that can be easily
        converted to JSON, YAML or other formats.

        .. code-block:: python

            [
                {
                    'id': 'test',
                    'roles': {
                        'client': 1,
                        'ldap': 1
                }
            ]

        :rtype: dict
        """

        out = []
        for domain in self.domains:
            out.append(domain.export())

        return out

    def satisfies(self, other: "Topology") -> bool:
        """
        Check if the topology satisfies the ``other`` topology.

        Returns ``True`` if this topology contains all domains and required
        roles defined in the ``other`` topology and ``False`` otherwise.

        :param other: The other topology.
        :type other: Topology
        :rtype: bool
        """
        for domain in other.domains:
            if domain.id not in self:
                return False

            if not self.get(domain.id).satisfies(domain):
                return False

        return True

    def __str__(self) -> str:
        return str(self.export())

    def __contains__(self, item: str) -> bool:
        try:
            return self.get(item) is not None
        except KeyError:
            return False

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.export() == other.export()

    def __ne__(self, other: object) -> bool:
        return not self == other

    @classmethod
    def FromMultihostConfig(cls, mhc: dict) -> "Topology":
        """
        Create :class:`Topology` from multihost configuration object.

        :param mhc: Multihost configuration object (dictionary)
        :type mhc: dict
        :return: Inferred topology.
        :rtype: Topology
        """

        if mhc is None:
            return cls()

        topology = []
        for domain in mhc.get("domains", []):
            topology.append({"id": domain["id"], "hosts": dict(Counter([x["role"] for x in domain["hosts"]]))})

        domains = [TopologyDomain(x.get("id", "default"), **x["hosts"]) for x in topology]
        return cls(*domains)



from __future__ import annotations

from .. import MultihostHost, MultihostUtility
from ..ssh import SSHLog, SSHProcess, SSHProcessResult

__all__ = ["SystemdServices"]


class SystemdServices(MultihostUtility):
    """
    Manage remote services.
    """

    def __init__(self, host: MultihostHost) -> None:
        super().__init__(host)
        self.initial_states: dict[str, bool] = {}

    def teardown(self) -> None:
        # Restart all services that were touched
        self.reload_daemon()
        for service, state in self.initial_states.items():
            self.host.ssh.run(
                f'systemctl stop "{service}" || systemctl status "{service}"',
                raise_on_error=False,
                log_level=SSHLog.Error,
            )
            if state:
                self.host.ssh.run(
                    f'systemctl start "{service}" || systemctl status "{service}"',
                    raise_on_error=False,
                    log_level=SSHLog.Error,
                )

    def async_start(self, service: str) -> SSHProcess:
        """
        Start a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be started. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: SSHProcess
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: starting "{service}" asynchronously')
        return self.host.ssh.async_run(
            f'systemctl start "{service}" || systemctl status "{service}"', log_level=SSHLog.Error
        )

    def start(self, service: str, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Start a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be started. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: starting "{service}"')
        return self.host.ssh.run(
            f'systemctl start "{service}" || systemctl status "{service}"',
            raise_on_error=raise_on_error,
            log_level=SSHLog.Error,
        )

    def async_stop(self, service: str) -> SSHProcess:
        """
        Stop a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be stopped. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: SSHProcess
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: stopping "{service}" asynchronously')
        return self.host.ssh.async_run(
            f'systemctl stop "{service}" || systemctl status "{service}"', log_level=SSHLog.Error
        )

    def stop(self, service: str, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Stop a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be stoped. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: stopping "{service}"')
        return self.host.ssh.run(
            f'systemctl stop "{service}" || systemctl status "{service}"',
            raise_on_error=raise_on_error,
            log_level=SSHLog.Error,
        )

    def async_restart(self, service: str) -> SSHProcess:
        """
        Restart a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be restarted. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: SSHProcess
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: restarting "{service}" asynchronously')
        return self.host.ssh.async_run(
            f'systemctl restart "{service}" || systemctl status "{service}"', log_level=SSHLog.Error
        )

    def restart(self, service: str, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Restart a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be restarted. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        self.__set_initial_state(service)
        self.logger.info(f'systemd: restarting "{service}"')
        return self.host.ssh.run(
            f'systemctl restart "{service}" || systemctl status "{service}"',
            raise_on_error=raise_on_error,
            log_level=SSHLog.Error,
        )

    def async_reload(self, service: str) -> SSHProcess:
        """
        Reload a systemd unit. Non-blocking call.

        ``systemctl status $unit`` is called automatically if the unit can not
        be reloaded. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: SSHProcess
        """
        self.logger.info(f'systemd: reloading "{service}" asynchronously')
        return self.host.ssh.async_run(
            f'systemctl reload "{service}" || systemctl status "{service}"', log_level=SSHLog.Error
        )

    def reload(self, service: str, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Reload a systemd unit. The call will wait until the operation is finished.

        ``systemctl status $unit`` is called automatically if the unit can not
        be reloaded. The status is then visible in the logs.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        self.logger.info(f'systemd: reloading "{service}"')
        return self.host.ssh.run(
            f'systemctl reload "{service}" || systemctl status "{service}"',
            raise_on_error=raise_on_error,
            log_level=SSHLog.Error,
        )

    def async_status(self, service: str) -> SSHProcess:
        """
        Get systemd unit status. Non-blocking call.

        :param service: Unit name.
        :type service: str
        :return: Running SSH process.
        :rtype: SSHProcess
        """
        return self.host.ssh.async_run(f'systemctl status "{service}"')

    def status(self, service: str, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Get systemd unit status. The call will wait until the operation is finished.

        :param service: Unit name.
        :type service: str
        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        return self.host.ssh.run(f'systemctl status "{service}"', raise_on_error=raise_on_error)

    def async_reload_daemon(self) -> SSHProcess:
        """
        Reload systemd daemon to refresh unit files. Non-blocking call.

        :return: Running SSH process.
        :rtype: SSHProcess
        """
        self.logger.info("systemd: reloading systemd daemon")
        return self.host.ssh.async_run("systemctl daemon-reload", log_level=SSHLog.Error)

    def reload_daemon(self, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Reload systemd daemon to refresh unit files. The call will wait until the operation is finished.

        :param raise_on_error: Raise exception on error, defaults to True
        :type raise_on_error: bool, optional
        :return: SSH process result.
        :rtype: SSHProcessResult
        """
        return self.host.ssh.run("systemctl daemon-reload", raise_on_error=raise_on_error)

    def __set_initial_state(self, service: str) -> None:
        if service in self.initial_states:
            return

        result = self.host.ssh.run(f'systemctl status "{service}"', log_level=SSHLog.Silent, raise_on_error=False)
        self.initial_states[service] = result.rc == 0



from __future__ import annotations

import itertools
import os
import shlex
import textwrap
from enum import Enum, auto
from typing import Any, Generator, Type

import colorama as c
import pssh.clients.base.single
import pssh.clients.ssh
import pssh.exceptions
import pssh.output

from ._private.logging import MultihostLogger


class SSHLog(Enum):
    """
    SSH command log level.
    """

    Silent = auto()
    """
    No log messages are produced.
    """

    Short = auto()
    """
    Command execution and return code is logged. Its output is omitted.
    """

    Full = auto()
    """
    Command execution, its return code and output is logged.
    """

    Error = auto()
    """
    Only log the command and its result on non-zero exit code.
    """


class SSHProcess(object):
    """
    SSH Process.

    .. note::

        You should not create instances of this class yourself. Use method
        :meth:`SSHClient.run`, :meth:`SSHClient.exec`,
        :meth:`SSHClient.async_run` and :meth:`SSHClient.async_exec` from
        :class:`SSHClient` to execute a command over SSH.
    """

    __genid = itertools.count()

    def __init__(
        self,
        *,
        command: str,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        shell: str | None = None,
        conn: pssh.clients.ssh.SSHClient,
        read_timeout: float,
        logger: MultihostLogger,
        log_level: SSHLog,
        sync_exec: bool,
    ) -> None:
        """
        :param command: Command to execute.
        :type command: str
        :param cwd: Working directory, defaults to None
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param shell: Shell used to execute the command, defaults to None (use user's login shell)
        :type shell: str | None, optional
        :param conn: Connected SSH client.
        :type conn: pssh.clients.ssh.SSHClient
        :param read_timeout: Timeout in seconds, how long should the client wait
            for output, defaults to 30 seconds
        :type read_timeout: float
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param log_level: Log level.
        :type log_level: SSHLog
        :param sync_exec: Is this a blocking execution?
        :type sync_exec: bool
        """
        self.__conn: pssh.clients.ssh.SSHClient = conn
        self.__process: pssh.output.HostOutput | None = None

        self.__logger: MultihostLogger = logger
        self.__log_level: SSHLog = self._get_log_level(log_level)
        self.__sync_exec: bool = sync_exec

        self.id = next(self.__genid) + 1
        self.command: str = textwrap.dedent(command).strip()
        self.cwd: str | None = cwd
        self.env: dict[str, Any] = env if env is not None else {}
        self.input: str | None = input
        self.shell: str | None = shell
        self.read_timeout: float = read_timeout

        self.__stdout_generator: Generator[str, None, None] | None = None
        self.__stderr_generator: Generator[str, None, None] | None = None
        self.__stdout: list[str] = []
        self.__stderr: list[str] = []

    @property
    def stdout(self) -> Generator[str, None, None]:
        """
        Standard output, returns generator which yields output line by line.

        .. code-block:: python

            # Read single line, this will block until there is a line to read or read_timeout is reached
            line = next(process.stdout)

            # Read all lines, this will block until EOF or read_timeout is reached
            lines = list(process.stdout)

            # Iterate over all lines
            for line in process.stdout:
                pass

        :raises RuntimeError: If the process is not yet started.
        :return: Standard output generator.
        :rtype: Generator[str, None, None]
        """
        if self.__stdout_generator is None:
            raise RuntimeError("The process has not yet started")

        return self.__stdout_generator

    @property
    def stderr(self) -> Generator[str, None, None]:
        """
        Standard error output, returns generator which yields error output line by line.

        .. code-block:: python

            # Read single line, this will block until there is a line to read or read_timeout is reached
            line = next(process.stderr)

            # Read all lines, this will block until EOF or read_timeout is reached
            lines = list(process.stderr)

            # Iterate over all lines
            for line in process.stderr:
                pass

        :raises RuntimeError: If the process is not yet started.
        :return: Standard error output generator.
        :rtype: Generator[str, None, None]
        """
        if self.__stderr_generator is None:
            raise RuntimeError("The process has not yet started")

        return self.__stderr_generator

    @property
    def stdin(self) -> pssh.clients.base.single.Stdin:
        """
        File-like object representing command's standard input.

        .. code-block:: python

            # Write data
            process.stdin.write('Hello World')

            # Send EOF to indicate that there will be no more input data.
            process.send_eof()

        :raises RuntimeError: If the process is not yet started.
        :return: Standard input file.
        :rtype: pssh.clients.base.single.Stdin
        """

        if self.__process is None:
            raise RuntimeError("The process has not yet started")

        return self.__process.stdin

    def run(self) -> SSHProcess:
        """
        Execute the command.

        :return: Self.
        :rtype: SSHProcess
        """
        complete_command = self._escape_command(self._build_complete_command(self.command, cwd=self.cwd, env=self.env))

        if self.__log_level in (SSHLog.Short, SSHLog.Full):
            self.__logger.info(
                self.__msg_execution(),
                extra={
                    "data": {
                        "Host": self.__conn.host,
                        "Shell": self.shell,
                        "User": self.__conn.user,
                        "Command": self.command,
                        "Input": self.input,
                        "Working directory": self.cwd,
                        "Extra environment": self.env,
                    }
                },
            )

        self.__process = self.__conn.run_command(
            command=complete_command,
            shell=self.shell,
            read_timeout=self.read_timeout,
        )

        def wrap_generator(generator, buffer) -> Generator[str, None, None]:
            for line in generator:
                buffer.append(line)
                yield line

        self.__stdout_generator = wrap_generator(self.__process.stdout, self.__stdout)
        self.__stderr_generator = wrap_generator(self.__process.stderr, self.__stderr)

        if self.input is not None:
            self.stdin.write(self.input)

        return self

    def wait(self, raise_on_error: bool = True) -> SSHProcessResult:
        """
        Wait for the command to finish.

        EOF is send to standard input to indicate that there will be no
        additional input data. Then it waits for the command to finish.

        :param raise_on_error: If True, :class:`SSHProcessError` is raised on non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :raises SSHProcessError: If ``raise_on_error`` is True and the command exited with non-zero return code.
        :return: Command result.
        :rtype: SSHProcessResult
        """
        if self.__process is None:
            raise RuntimeError("Calling wait on process that has not yet started.")

        self.send_eof()
        self.__conn.wait_finished(self.__process)

        # Read remaining output, this will finish the output generator and append
        # remaining lines to self.__stdout and self.__stderr buffers.
        list(self.stdout)
        list(self.stderr)

        # Get exit code.
        code = self.__conn._eagain_errcode(self.__process.channel.get_exit_status, -1)

        result = SSHProcessResult(code, self.__stdout, self.__stderr)

        if self.__log_level == SSHLog.Error and result.rc != 0:
            self.__logger.error(
                self.__msg_completed_async(result.rc),
                extra={
                    "data": {
                        "Host": self.__conn.host,
                        "User": self.__conn.user,
                        "Command": self.command,
                        "Input": self.input,
                        "Working directory": self.cwd,
                        "Extra environment": self.env,
                        "Output": result.stdout,
                        "Error output": result.stderr,
                    }
                },
            )

        if self.__sync_exec:
            match self.__log_level:
                case SSHLog.Short:
                    self.__logger.info(self.__msg_completed_sync(result.rc))
                case SSHLog.Full:
                    self.__logger.info(
                        self.__msg_completed_sync(result.rc),
                        extra={
                            "data": {
                                "Output": result.stdout,
                                "Error output": result.stderr,
                            }
                        },
                    )
                case _:
                    pass
        else:
            match self.__log_level:
                case SSHLog.Short:
                    self.__logger.info(
                        self.__msg_completed_async(result.rc),
                        extra={
                            "data": {
                                "Host": self.__conn.host,
                                "User": self.__conn.user,
                                "Command": self.command,
                                "Input": self.input,
                                "Working directory": self.cwd,
                                "Extra environment": self.env,
                            }
                        },
                    )
                case SSHLog.Full:
                    self.__logger.info(
                        self.__msg_completed_async(result.rc),
                        extra={
                            "data": {
                                "Host": self.__conn.host,
                                "User": self.__conn.user,
                                "Command": self.command,
                                "Input": self.input,
                                "Working directory": self.cwd,
                                "Extra environment": self.env,
                                "Output": result.stdout,
                                "Error output": result.stderr,
                            }
                        },
                    )
                case _:
                    pass

        if raise_on_error and result.rc != 0:
            raise SSHProcessError(
                self.id, self.command, result.rc, self.cwd, self.env, self.input, result.stdout, result.stderr
            )

        return result

    def send_eof(self) -> None:
        """
        Send EOF to standard input to indicate that there will be no more
        input data.

        :raises RuntimeError: If the process is not yet started.
        """
        if self.__process is None:
            raise RuntimeError("The process has not yet started")

        self.__process.channel.send_eof()

    def _build_complete_command(self, command: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        out = ""

        # Set environment variables
        for key, value in env.items():
            out += f"export {key}={shlex.quote(str(value))}\n"

        # Set working directory
        if cwd is not None:
            out += f"cd {shlex.quote(cwd)}\n"

        if out:
            out += "\n"

        out += command

        return out

    def _escape_command(self, command: str) -> str:
        """
        pssh simply calls the command as $shell '$command', e.g.
        bash -c '$command'

        We need to escape ' inside the command to make it work correctly.
        """
        return command.replace("'", "'\"'\"'")

    def _get_log_level(self, input_log_level: SSHLog) -> SSHLog:
        debug = os.getenv("MH_SSH_DEBUG", "no")
        if debug.lower() in ["true", "yes", "1"]:
            return SSHLog.Full

        return input_log_level

    def __msg_id(self) -> str:
        return self.__logger.colorize(f"#{self.id}", c.Style.BRIGHT, c.Fore.BLUE)

    def __msg_rc(self, rc: int) -> str:
        if rc == 0:
            return self.__logger.colorize(rc, c.Style.BRIGHT, c.Fore.GREEN)

        return self.__logger.colorize(rc, c.Style.BRIGHT, c.Fore.RED)

    def __msg_execution(self) -> str:
        return f'{self.__logger.colorize("Executing command", c.Style.BRIGHT)} ' + self.__msg_id()

    def __msg_completed_sync(self, rc: int) -> str:
        return "Previous command completed with exit code " + self.__msg_rc(rc)

    def __msg_completed_async(self, rc: int) -> str:
        return (
            self.__logger.colorize("Command ", c.Style.BRIGHT)
            + self.__msg_id()
            + self.__logger.colorize(" completed with exit code ", c.Style.BRIGHT)
            + self.__msg_rc(rc)
        )

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.wait()


class SSHBashProcess(SSHProcess):
    """
    SSH Process with Bash.

    .. note::

        You should not create instances of this class yourself. Use method
        :meth:`SSHClient.run`, :meth:`SSHClient.exec`,
        :meth:`SSHClient.async_run` and :meth:`SSHClient.async_exec` from
        :class:`SSHClient` to execute a command over SSH.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, shell="/usr/bin/bash -c")


class SSHPowerShellProcess(SSHProcess):
    """
    SSH Process with Powershell.

    .. note::

        You should not create instances of this class yourself. Use method
        :meth:`SSHClient.run`, :meth:`SSHClient.exec`,
        :meth:`SSHClient.async_run` and :meth:`SSHClient.async_exec` from
        :class:`SSHClient` to execute a command over SSH.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, shell="powershell -NonInteractive -Command")

    def _build_complete_command(self, command: str, *, cwd: str | None, env: dict[str, Any]) -> str:
        out = ""

        # Set environment variables
        for key, value in env.items():
            out += f"$Env:{key} = {shlex.quote(str(value))}\n"

        # Set working directory
        if cwd is not None:
            out += f"cd {shlex.quote(cwd)}\n"

        if out:
            out += "\n"

        out += command

        return out

    def _escape_command(self, command: str) -> str:
        """
        pssh simply calls the command as $shell '$command', e.g.
        bash -c '$command'

        We need to escape ' inside the command to make it work correctly.
        """
        return command.replace("'", "''").replace('"', '\\"')


class SSHProcessResult(object):
    """
    SSH Process result.
    """

    def __init__(self, rc: int, stdout: list[str], stderr: list[str]) -> None:
        """
        :param rc: Return code.
        :type rc: int
        :param stdout: Standard output, line by line.
        :type stdout: list[str]
        :param stderr: Standard error output, line by line.
        :type stderr: list[str]
        """
        self.rc = rc
        self.stdout: str = "\n".join(stdout)
        self.stderr: str = "\n".join(stderr)
        self.stdout_lines: list[str] = stdout
        self.stderr_lines: list[str] = stderr


class SSHProcessError(Exception):
    """
    SSH Process Error.
    """

    def __init__(
        self,
        id: int,
        command: str,
        rc: int,
        cwd: str | None,
        env: dict[str, Any],
        input: str | None,
        stdout: str,
        stderr: str,
    ) -> None:
        pretty_env = ""
        for key, value in env.items():
            pretty_env += f"{key}={value}\n"

        def dumps(value) -> str:
            if not value:
                return ""

            return "\n" + textwrap.indent(value, " " * 12)

        super().__init__(
            textwrap.dedent(
                f"""
        Command #{id} exited with return code {rc}:
          Command:{dumps(command)}
          CWD:{dumps(cwd)}
          Env:{dumps(pretty_env.strip())}
          Output:{dumps(stdout)}
          Error output:{dumps(stderr)}
        """
            )
        )

        self.id = id
        self.command = (command,)
        self.rc = rc
        self.cwd = cwd
        self.env = env
        self.input = input
        self.stdout = stdout
        self.stderr = stderr


class SSHAuthenticationError(Exception):
    def __init__(
        self,
        host: str,
        port: int,
        user: str,
    ) -> None:
        super().__init__(f'Unable to authenticate as "{user}" at {host}:{port} over SSH')


class SSHClient(object):
    """
    Interactive SSH client.

    .. code-block:: python
        :caption: Example: Blocking call

        # Connect to SSH server, it is automatically disconnected when leaving the with statement
        with SSHClient(host, user=username, password=password, logger=logger) as ssh:
            result = ssh.run('echo Hello World')
            print(result.rc)
            print(result.stdout)

            result = ssh.run('cat', input='Hello World')
            print(result.rc)
            print(result.stdout)

    .. code-block:: python
        :caption: Example: Non-blocking call

        # Connect to SSH server, it is automatically disconnected when leaving the with statement
        with SSHClient(host, user=username, password=password, logger=logger) as ssh:
            # The process is executed, but it does not block. In order to wait for it to finish, run process.wait()
            process = ssh.async_run('echo Hello World')
            result = process.wait()
            print(result.rc)
            print(result.stdout)

            # You can write to stdin directly in asynchronous run
            process = ssh.async_run('cat')
            process.stdin.write('Hello World')
            process.send_eof()
            result = process.wait()
            print(result.rc)
            print(result.stdout)

            # You can also work with inputs and outputs more interactively.
            # The process is automatically waited when leaving the with statement.
            with ssh.async_run('bash') as process:
                process.stdin.write('echo Hello World\\n')
                print(next(process.stdout))

                process.stdin.write('echo This works as well\\n')
                print(next(process.stdout))

    .. note::

        It is possible to set ``MH_SSH_DEBUG=yes`` environment variable to
        log output and exist status to from commands, regardless of what log
        level is used. This essentially enforces the :attr:`SSHLog.Full` level.
    """

    def __init__(
        self,
        host: str,
        *,
        user: str,
        password: str,
        port: int = 22,
        shell: Type[SSHProcess] = SSHProcess,
        logger: MultihostLogger,
    ) -> None:
        """
        :param host: Host name to connect to.
        :type host: BaseRole | str
        :param user: Username to authenticate.
        :type user: str
        :param password: Password.
        :type password: str
        :param logger: Multihost logger.
        :type logger: MultihostLogger
        :param port: SSH port, defaults to 22
        :type port: int, optional
        :param shell: User shell used to run commands, defaults to '/usr/bin/bash -c'
        :type shell: str, optional
        """
        self.host: str = host
        self.user: str = user
        self.password: str = password
        self.port: int = port
        self.shell: Type[SSHProcess] = shell
        self.logger: MultihostLogger = logger

        self.__conn: pssh.clients.ssh.SSHClient | None = None

    @property
    def connected(self) -> bool:
        """
        :return: True if the client is connected, False otherwise.
        :rtype: bool
        """
        return self.__conn is not None

    @property
    def conn(self) -> pssh.clients.ssh.SSHClient:
        """
        Low-level connection object.

        :return: Parallel-ssh connection object.
        :rtype: pssh.clients.ssh.SSHClient
        """
        if self.__conn is None:
            RuntimeError("SSH client is not connected.")

        return self.__conn

    def connect(self) -> None:
        """
        Connect to the host.

        :raises SSHAuthenticationError: If user fails to authenticate.
        """
        if self.connected:
            return

        self.logger.info(
            self.logger.colorize("Opening SSH connection to ", c.Style.BRIGHT)
            + self.logger.colorize(self.host, c.Fore.BLUE, c.Style.BRIGHT)
        )

        try:
            self.__conn = pssh.clients.ssh.SSHClient(
                host=self.host,
                user=self.user,
                password=self.password,
                port=self.port,
                identity_auth=False,
                gssapi_auth=False,
                allow_agent=False,
                num_retries=1,
            )
        except pssh.exceptions.AuthenticationError:
            e = SSHAuthenticationError(self.host, self.port, self.user)
            self.logger.error(str(e))
            raise e

    def disconnect(self) -> None:
        """
        Disconnect client.
        """
        self.logger.info(
            self.logger.colorize("Closing SSH connection to ", c.Style.BRIGHT)
            + self.logger.colorize(self.host, c.Fore.BLUE, c.Style.BRIGHT)
        )

        if self.__conn is None:
            return

        self.__conn.disconnect()
        self.__conn = None

    def async_run(
        self,
        command: str,
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        read_timeout: float = 30,
        log_level: SSHLog = SSHLog.Full,
    ) -> SSHProcess:
        """
        Non-blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately, however it does not wait for the command to finish.

        :param command: Command to run.
        :type command: str
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param read_timeout: Timeout in seconds, how long should the client wait for output, defaults to 30 seconds
        :type read_timeout: float, optional
        :param log_level: Log level, defaults to SSHLog.Full
        :type log_level: SSHLog, optional
        :return: Instance of :class:`SSHProcess`, the process is already running.
        :rtype: SSHProcess
        """
        if not isinstance(command, str):
            raise ValueError("Parameter command is not a string, did you mean async_exec() instead of async_run()?")

        self.connect()

        process = self.shell(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            conn=self.conn,
            read_timeout=read_timeout,
            logger=self.logger,
            log_level=log_level,
            sync_exec=False,
        )

        process.run()
        return process

    def run(
        self,
        command: str,
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        read_timeout: float = 2,
        log_level: SSHLog = SSHLog.Full,
        raise_on_error: bool = True,
    ) -> SSHProcessResult:
        """
        Blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately. It waits for the command to finish and returns
        its result.

        :param command: Command to run.
        :type command: str
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param read_timeout: Timeout in seconds, how long should the client wait
            for output, defaults to 30 seconds
        :type read_timeout: float, optional
        :param log_level: Log level, defaults to SSHLog.Full
        :type log_level: SSHLog, optional
        :param raise_on_error: If True, raise :class:`SSHProcessError` if
            command exited with non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :raises SSHProcessError: If ``raise_on_error`` is True and the command exited with non-zero return code.
        :return: Command result.
        :rtype: SSHProcessResult
        """
        if not isinstance(command, str):
            raise ValueError("Parameter command is not a string, did you mean exec() instead of run()?")

        self.connect()

        process = self.shell(
            command=command,
            cwd=cwd,
            env=env,
            input=input,
            conn=self.conn,
            read_timeout=read_timeout,
            logger=self.logger,
            log_level=log_level,
            sync_exec=True,
        )

        process.run()

        return process.wait(raise_on_error=raise_on_error)

    def async_exec(
        self,
        argv: list[Any],
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        read_timeout: float = 2,
        log_level: SSHLog = SSHLog.Full,
    ) -> SSHProcess:
        """
        Non-blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately, however it does not wait for the command to finish.

        The command is provided as ``argv`` list.

        :param argv: Command to run.
        :type argv: list[Any]
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param read_timeout: Timeout in seconds, how long should the client wait for output, defaults to 30 seconds
        :type read_timeout: float, optional
        :param log_level: Log level, defaults to SSHLog.Full
        :type log_level: SSHLog, optional
        :return: Instance of :class:`SSHProcess`, the process is already running.
        :rtype: SSHProcess
        """
        if not isinstance(argv, list):
            raise ValueError("Parameter argv is not a list, did you mean async_run() instead of async_exec()?")

        argv = [str(x) for x in argv]
        command = shlex.join(argv)

        return self.async_run(
            command,
            cwd=cwd,
            env=env,
            input=input,
            read_timeout=read_timeout,
            log_level=log_level,
        )

    def exec(
        self,
        argv: list[Any],
        *,
        cwd: str | None = None,
        env: dict[str, Any] | None = None,
        input: str | None = None,
        read_timeout: float = 2,
        log_level: SSHLog = SSHLog.Full,
        raise_on_error: bool = True,
    ) -> SSHProcessResult:
        """
        Blocking command call.

        The command is run under shell specified in the constructor and it is
        executed immediately. It waits for the command to finish and returns its
        result.

        The command is provided as ``argv`` list.

        :param argv: Command to run.
        :type argv: list[Any]
        :param cwd: Working directory, defaults to None (= do not change)
        :type cwd: str | None, optional
        :param env: Additional environment variables, defaults to None
        :type env: dict[str, Any] | None, optional
        :param input: Content of standard input, defaults to None
        :type input: str | None, optional
        :param read_timeout: Timeout in seconds, how long should the client wait
            for output, defaults to 30 seconds
        :type read_timeout: float, optional
        :param log_level: Log level, defaults to SSHLog.Full
        :type log_level: SSHLog, optional
        :param raise_on_error: If True, raise :class:`SSHProcessError` if
            command exited with non-zero return code, defaults to True
        :type raise_on_error: bool, optional
        :raises SSHProcessError: If ``raise_on_error`` is True and the command exited with non-zero return code.
        :return: Command result.
        :rtype: SSHProcessResult
        """
        if not isinstance(argv, list):
            raise ValueError("Parameter argv is not a list, did you mean run() instead of exec()?")

        argv = [str(x) for x in argv]
        command = shlex.join(argv)

        return self.run(
            command,
            cwd=cwd,
            env=env,
            input=input,
            read_timeout=read_timeout,
            log_level=log_level,
            raise_on_error=raise_on_error,
        )

    def expect(self, expect_script: str, raise_on_error: bool = False) -> SSHProcessResult:
        """
        Run expect script.

        :param expect_script: Expect script.
        :type expect_script: str
        :param raise_on_error: If True, raise :class:`SSHProcessError` if
            command exited with non-zero return code, defaults to False
        :type raise_on_error: bool, optional
        :return: Expect script result.
        :rtype: SSHProcessResult
        """
        return self.run("/bin/expect -d", input=expect_script, raise_on_error=raise_on_error)

    def expect_nobody(self, expect_script: str, raise_on_error: bool = False) -> SSHProcessResult:
        """
        Run expect script as user nobody.

        The main use case is to avoid running the command as root if the client
        is connected to the root user SSH session.

        :param expect_script: Expect script.
        :type expect_script: str
        :param raise_on_error: If True, raise :class:`SSHProcessError` if
            command exited with non-zero return code, defaults to False
        :type raise_on_error: bool, optional
        :return: Expect return code.
        :rtype: SSHProcessResult
        """
        return self.run(
            'su --shell /bin/sh nobody -c "/bin/expect -d"', input=expect_script, raise_on_error=raise_on_error
        )

    def __enter__(self) -> SSHClient:
        """
        Connect to the host.

        :return: SSHClient instance.
        :rtype: SSHClient
        """
        self.connect()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        """
        Disconnect.
        """
        self.disconnect()



from __future__ import annotations

from typing import Any


def validate_configuration(
    required_keys: list[str], confdict: dict[str, Any], error_fmt: str = '"{key}" property is missing'
) -> None:
    """
    Validate configuration dictionary.

    Check that it contains all required keys. The key may contain ``.`` to check
    nested keys, for example ``ssh.user``.

    :param required_keys: Required keys.
    :type required_keys: list[str]
    :param confdict: Configuration dictionary.
    :type confdict: dict[str, Any]
    :param error_fmt: _description_, defaults to '"{key}" property is missing'
    :type error_fmt: str, optional
    :raises ValueError: If a required key is missing or empty.
    :return: ``True`` if all keys are present and not empty, ``False`` otherwise.
    :rtype: bool
    """

    def is_property_in_dict(property: str, d: dict[str, Any]) -> bool:
        if "." in property:
            (key, subpath) = property.split(".", maxsplit=1)
            if not d.get(key, None):
                return False

            return is_property_in_dict(subpath, d[key])

        return property in d and d[property]

    for key in required_keys:
        if not is_property_in_dict(key, confdict):
            raise ValueError(error_fmt.format(key=key))
